# -*- coding: utf-8 -*-
"""
Created on Fri Sep 15 13:00:08 2023

@author: anba
"""

import numpy as np 
import pandas as pd 
import os, sys
import pyproj
# from pyproj import Transformer
import copy
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from matplotlib.collections import LineCollection
import matplotlib.ticker as mticker
import cmocean
import warnings

import rich
import dill
import copy

#import pyplume as pp
from ..ptools import ptools
from ..plotting.matplotlib_shell import subplots, dhi_colors


class Pose:
    """
    Class for managing monitoring platform pose and orientation ("Pose") data (e.g. ROV, Vessel, AUV, Lander, etc).
    positive z is up, positive x is east, and positive y is north. 
    Features:
    - Importing data (pose and orientation)
    - Regularization/re-sampling
    - Updating pose depth (z)
    - QAQC when possible
    - Reversing z-convention
    - Scaling operations (e.g., vertical exaggeration)
    - managing projections 
    """

    def __init__(self, name='Pose', df=None, z_convention = 'normal', proj = "EPSG:32611"):
        """
        Initialize the Pose class.

        Parameters
        ----------
        name : str, optional
            Name of the pose data. Default is 'Pose'.
        df : pd.DataFrame, optional
            Input pandas DataFrame. Default is None.
        z_convention : str, optional
            Specify the z-convention of the data as 'normal' (positive z is up) or 'reverse' (positive z is down). Default is 'normal'.
        proj: str, optional
            data projection. Use pyproj projection keys. 
        """
        self.pose = None
        self.name = name
        self.plot = self.__plotting(self)
        
        self.resample_method = 'nearest'
        self.resample_tolerance = None #second
        
        
            
        self.proj = proj
        


        if df is not None:
            self.import_from_pandas_df(df)
            
        # set z convention
        # doesnt work with the funciton call on init for some reason 
        if z_convention in ['normal','reverse']:
            self.z_convention = z_convention # positive z is up, negative z is down
        else:
            print(f'invalid z_convention. Choose from {z_convention}')    

        
    def set_resample_method(self,method): 
        self.resample_method = method
        
    def set_resample_tolerance(self,tolerance):
        # if method is nearest this is the tolerance for matchingh
        # if ffill of bfill its the consecutive numberto propageate
        self.resample_tolerance = tolerance
        
    def to_csv(self, fname=None):
        """
        Write pose data to a CSV file.

        Parameters
        ----------
        fname : str, optional
            The file name to save the CSV. If not provided, the default filename
            will be based on the 'name' attribute of the pose instance.
        """
        if not fname:
            fname = f'{self.name}_pose_export.csv'
        self.pose.to_csv(fname)

    def import_from_pandas_df(self, df):
        """
        Import data from a pandas DataFrame and transform column names.

        DataFrame column names must include 'Easting', 'Northing', 'Depth',
        'Pitch', 'Roll', and 'Heading'.

        Parameters
        ----------
        df : pd.DataFrame
            The input pandas DataFrame.
        flip_z : bool, optional
            Whether to flip the z-values. Default is False.

        Returns
        -------
        pd.DataFrame
            Transformed DataFrame with updated column names.
        """
        self.pose = df[['Easting', 'Northing', 'Depth', 'Pitch', 'Roll', 'Heading']].copy()
        self.pose.columns = ['x', 'y', 'z', 'pitch', 'roll', 'heading']
        self.pose['z'] = -self.pose['z']
        
        # if flip_z:
        #     self.pose['z'] = -self.pose['z']  # Flip z direction so that down is negative
    #@z_convention.setter    
    def set_z_convention(self,z_convention = 'normal', update_z = False):
        """
        set the z-coordinate convention of the pose data.

        This method multiplies the z-coordinates by -1 to flip the convention,
        ensuring that down is negative.
        
        Parameters
        ----------
        z_convention : str, optional
            Specify the z-convention of the data as 'normal' (positive z is up) or 'reverse' (positive z is down). Default is 'normal'.
            
        update_z: bool, optional
            if true the z-coordinate will be multiplied by -1. Default is False. 
        Notes
        -----
        This method modifies the 'z' column of the 'pos' DataFrame in place.
        """
        
        if z_convention in ['normal','reverse']:
            self.z_convention = z_convention # positive z is up, negative z is down
        else:
            print(f'invalid z_convention. Choose from {z_convention}')
            
            #raise Warning(f'invalid z_convention. Choose from {z_convention}')
           
        if update_z:
            self.pose['z'] = -self.pose['z']

    def get_between(self, start_date, end_date):
        """
        Retrieve pose data between two dates.

        Parameters
        ----------
        start_date : datetime-like
            The start date for filtering the pose data.
        end_date : datetime-like
            The end date for filtering the pose data.

        Returns
        -------
        pandas.DataFrame
            pose data within the specified date range.

        Notes
        -----
        The 'pose' DataFrame must have a DatetimeIndex for this method to work correctly.
        """
        mask = (self.pose.index >= start_date) & (self.pose.index <= end_date)
        return self.pose.loc[mask]
    
    

            

            
            # for i,etime in enumerate(et):
            #     index = orient_in.index.get_indexer([etime], method='ffill')[0] # index of nearest input positon to the ensemble timestamp
            #     self.orientation[0,i] = orient_in.iloc[index].pitch
            #     self.orientation[1,i] = orient_in.iloc[index].roll
            #     self.orientation[2,i] = orient_in.iloc[index].heading 
    def add_lat_lon(self):
        """
        Convert from input easting northing to lat/lon.
        
        Input projection code comes from pyproj.
        Lat and lon columns are added to the pose data.
        """
    
        transformer = pyproj.Transformer.from_crs(self.proj, "EPSG:4326")
        lat, lon = transformer.transform(self.pose['x'].to_numpy(), self.pose['y'].to_numpy())
        
        self.pose['latitude'] = lat
        self.pose['longitude'] = lon

        
        
        
            
    class __plotting:
        def __init__(self, pose):
            self._pose = pose

        def trajectory(self, **kwargs):
            """
            Create a trajectory plot.

            Parameters
            ----------
            **kwargs : dict, optional
                Additional keyword arguments.

            Notes
            -----
            This method uses the 'pos' DataFrame stored in the pose class.
            """
            pose = self._pose.pose

            start_date = kwargs.pop('start_date', pose.index.min())
            end_date = kwargs.pop('end_date', pose.index.max())

            mask = (pose.index >= start_date) & (pose.index <= end_date)
            pose = pose.loc[mask]

            fig, ax = subplots(nrow = 3, figheight = 8, figwidth = 10)
            
            
            

            
            s = ax[0].scatter(pose['x'], pose['y'],s = 5, c = pose['z'], cmap = 'jet')
            ax[0].set_ylabel('Distance (m)')
            ax[0].set_aspect('equal')
            ax[0].grid(alpha = 0.3, color = 'white')
            ax[0].set_title('Position')
            fig.colorbar(s,ax = ax[0],label = 'Depth (m)')
            
            
            ax[1].scatter(pose.index, pose['z'],s = 5,c = pose['z'], cmap = 'jet')
            ax[1].set_ylabel('Depth (m)')
            ax[1].grid(alpha = 0.3, color = 'white')

            ax[2].plot(pose.index, pose['pitch'],lw = 1,ls = '--',c ='black',alpha = 0.9,label = 'Pitch')
            ax[2].plot(pose.index, pose['roll'] ,lw = 2,ls = '-',c ='black',alpha = 0.9,label = 'Roll')
            ax[2].set_ylabel('degrees')
            ax[2].grid(alpha = 0.3, color = 'white')
            ax[2].legend()
            
            ax[0].set_facecolor('lightgray')
            ax[1].set_facecolor('lightgray')
            ax[2].set_facecolor('lightgray')
            title = f'{self._pose.name}'
            fig.suptitle(title)  
            
            
            return fig,ax
            
            
    def resample_to(self, t_in,kwargs):
        """
        Resample the pose data to match the input timeseries.

        This method resamples the pose data to align with the provided input timeseries (`t_in`).
        The `tolerance` parameter specifies the maximum allowable time difference in seconds for a match to be considered acceptable.
        Input timesteps that could not be matched will be filled in with nan values. 
        It uses nearest neighbor search to find the closest matching timestamps in the input timeseries.
        
        Warning - all data in the pose data outside the bounds of the input 
        timeseries will be deleted. 

        Args:
            t_in (list or numpy array, pandas datatime index): A list-like object containing pandas datetimes representing the target timeseries.
            
            tolerance (int, optional): The maximum time difference in seconds allowed for matching. only true if method = nearest
            Default is 1 second. if tolerance is None, then the nearest value 
            will be used, regardless of distance. this is useful for static 
            sensors without regular pose measurements. 




        """
        
        # st = np.nanmin(t_in)
        # et = np.
        self.pose = self.pose.reindex(t_in, method=self.resample_method, tolerance = self.resample_tolerance)
        
        # method = kwargs.pop('method','nearest')
        # tolerance = kwargs.pop('tolerance','auto')

        # if tolerance:
        #     if tolerance == 'auto': # calculate the right tolerance 
        #         if len(t_in)>1 and len(self.pose.index)>1:
        #             # calculate timestep for input timeseries
                   
        #             dt = np.diff(t_in).astype(float)/1e9 # time delta in seconds
                    
        #             dt = np.append(dt,dt[-1])
        #             dt_99_in = np.percentile(dt,99) # mask to identify data gaps 
                    
        #             # calculate timestep for existing timeseries
        #             t = self.pose.index.to_numpy()
        #             dt = np.diff(t,prepend = t[0],).astype(float)/1e9 # time delta in seconds
        #             dt_99 = np.percentile(dt,99) # mask to identify data gaps 
                    
        #             tolerance = 1.1*max(dt_99,dt_99_in)
        #         else:
        #             tolerance = 1e6 # set to 30 :)
                    
                    
                

            
        #     print(tolerance)
        #     self.pose = self.pose.reindex(t_in, tolerance=pd.to_timedelta(f'{tolerance} s'), method='nearest')
        # else:
        #     self.pose = self.pose.reindex(t_in, method='nearest')
            
        #rich.print(f"[light_green]({self.name}) successfully resampled on interval")
                
        #rich.print(f"[bold cyan]({self.name})[white] pose data successfully resampled\n{(len(self.name)+3)*' '}[white bold]interval:[white] {t_in[0].strftime('%D %H:%M')}-{t_in[-1].strftime('%D %H:%M')}\n{(len(self.name)+3)*' '}[white bold]% valid:[white] {(self.pose.isnull().sum()/(~self.pose.isnull()).sum()).mean()*100} ")
        
        
        #return p_out
        
        
    def resample_from(self,tser,field_name, tolerance = 'auto', z_convention = 'normal'):
        """
        set field in the pose data ('x','y','z','pitch','roll','heading') 
        from an external timeseries. 
        
        This method resamples the input timeseries data to align with the pose timesteps
        
        Input timesteps that could not be matched will be filled in with nan values. 
        It uses nearest neighbor search to find the closest matching timestamps in the input timeseries.
 
        
        input timeseries must be a pandas series object 
        
        z_convention: str, optional
           z convention for the input timeseries. 'normal' (positive z in up) or 'reverse' (negative z is up). 
        """
        
        if tolerance == 'auto': # calculate the right tolerance 
        
        
            
            # calculate timestep for input timeseries
            t = tser.index.to_numpy()
            dt = np.diff(t,prepend = t[0],).astype(float)/1e9 # time delta in seconds
            dt_99_in = np.percentile(dt,99) # mask to identify data gaps 
            
            # calculate timestep for existing timeseries
            t = self.pose.index.to_numpy()
            dt = np.diff(t,prepend = t[0],).astype(float)/1e9 # time delta in seconds
            dt_99 = np.percentile(dt,99) # mask to identify data gaps 
            
            tolerance = max(dt_99,dt_99_in)
            

            
        
        tser = tser.reindex(self.pose.index,tolerance=pd.to_timedelta(f'{tolerance} s'), method='nearest')
        if field_name in self.pose.columns: # check if input field_name is valid
        
            self.pose[field_name] = tser.to_numpy()
            # if z_convention == self.z_convention:
            #     self.pose[field_name] = tser.to_numpy()
            # else:
            #     self.pose[field_name] = -1*tser.to_numpy()
                
        
        else:
            raise SyntaxWarning(f'{field_name} is not a valid field \nField options: {", ".join(self.pose.columns)}')
            
            
    def save(self, fpath = None):
        # serialize a Pose robject 
        if fpath == None:
            fpath = '.' + os.sep + self.name + '.pose'
        else:
            fpath = fpath + '.pose'
            
            
        with open(fpath, "wb") as dill_file:
            # serialize the data manager object and save to file
            dill.dump(self, dill_file)
            
            
    def copy(self, name, deep = False):
        """
        Make a copy of the pose object.

        Parameters
        ----------
        deep : bool, optional
            If True, perform a deep copy. If False, perform a shallow copy.
            The default is False.
            
        name: str
            name of the copied Pose object (updates self.name)
            

        Returns
        -------
        Pose
            A copy of the pose object.
        """
        
        
        if deep:
            pout = copy.deepcopy(self)
        else:
            pout = copy.copy(self)
            
            pout.name = name
        return pout
    
    def merge(self,pose_in):
        None

#%%
###############################################################################
def concatenate(pose_in, name, z_convention='automatic'):
    """
    Concatenate multiple pose objects.

    This function concatenates multiple pose objects into a single pose object. It checks the z-convention of all
    input data and adjusts it as necessary.

    Parameters:
        pose_in (list): A list of pose objects to concatenate.
        name (str): The name of the concatenated pose object.
        z_convention (str, optional): The z-convention for the input data. If set to 'automatic' (default), the function
            will determine the most frequent z-convention among the input pose objects.

    Returns:
        Pose.Pose: A concatenated pose object.

    Example:
        # Create a list of pose objects
        pose_objects = [pose1, pose2, pose3]

        # Concatenate the pose objects into a new pose object named 'concatenated_pose'
        concatenated_pose = concatenate(pose_objects, name='concatenated_pose')
    """
    def most_frequent(List):
        return max(set(List), key=List.count)

    # Determine the most frequent z-convention if 'automatic' is selected
    if z_convention == 'automatic':
        z_convention = most_frequent([p.z_convention for p in pose_in])

    pose_dfs = []
    for pose in pose_in:
        if pose.z_convention == z_convention:
            pose_dfs.append(pose.pose)
        else:
            pose.set_z_convention(z_convention, update_z=True)
            pose_dfs.append(pose.pose)

    pose = pd.concat(pose_dfs)
    pose = pose[~pose.index.duplicated(keep='first')].copy()
    pose.sort_index(inplace=True)
    pose = pose.dropna()

    p = Pose(name=name, z_convention=z_convention)
    p.pose = pose

    return p

def load(fpath):
    # deserialize a Pose object 
    with open(fpath, "rb") as dill_file:
        # deserialize data manager object and reinitilize the print consoles
         p = dill.load(dill_file)    

             
    return p


# -*- coding: utf-8 -*-
"""
Created on Mon Jul 31 11:27:01 2023

@author: anba
"""

from rich import print

import matplotlib.pyplot as plt
import matplotlib as mpl
import numpy as np
import pandas as pd
import warnings
import os
import sys
import time 
import psutil


from rich.console import Console
from rich.text import Text
from rich.style import Style
from rich.theme import Theme


import time


# from ..ptools import ptools
# from ..plotting.matplotlib_shell import subplots, dhi_colors
# from ..pd0.pd0_decoder import DataSet as DataSet_pd0,PT3,Pd0
# from ..ctd.seabird_hex_decoder import seabird_ctd, DataSet as DataSet_ctd


np.seterr(divide='ignore', invalid='ignore')

import dill





from pyplume import ptools
from pyplume.plotting.matplotlib_shell import subplots, dhi_colors
from pyplume.pd0.pd0_decoder import DataSet as DataSet_pd0,PT3,Pd0
from pyplume.ctd.seabird_hex_decoder import seabird_ctd, DataSet as DataSet_ctd




#%%



class manager_base:
    def __init__(self,parser_name,name,fpaths,min_fsize = 0):
        """
        Parameters
        ----------
        parser_name : str
            parser to use. 'adcp', 'ctd', etc...
        name : str
            name of the dataset.
        fpaths : list
            list of filepaths [f1,f2,f3], or list of tuples containing all 
            filepaths needed for the parser [(f1a,f1b,..),(f2a,f2b,..),...]. 
            In general the main data file should be first.
            
        min_fsize : int, optional
            minimum filesize (in kb) to accept. Files smaller than this will be
            rejected. The default is 0.

        Returns
        -------
        None.

        """
 
        self.name = name # name of the dataset
        self.type = parser_name # name of the parser to use 
        self._filepaths_in = fpaths # list of all input filepaths 
        self._min_fsize = min_fsize # minimum filesize to load 
        
        
        ## check for duplicate input files 
        
        
        # theme = Theme({'screening file': 'gold3 italic',
        #                'loading_file': 'dark_cyan',
        #                'notification': 'yellow'})
        # self.console = Console(tab_size=4,
        #                        log_path = False,
        #                        theme = theme) # create a rich-text console 
        self.create_console()

        self._reason_failed = [] # reason why files could not be imported
        self._filepaths_failed = [] # full path to all files that were unsuccessfully imported
        self.filepaths = [] # full path to all files that were successfully imported
        self.file_start_times = []# start times for all files in the dataset
        self.file_end_times = [] # end times for all files in the dataset
        self.n_files = 0 # number of files in the dataset
        self.dataset_size = 0 # size of the entire data set (in kb)
        self.data = [] # data currently being held in memory 
        self.labels = []
        
        
        ## set parser type 
        if self.type == 'pd0':
            self._parser = self.__pd0_parser
            
        elif self.type == 'ctd':
            self._parser = self.__ctd_parser
            
        # elif self.type == 'pose':
        #     self._parser = self.__pose_parser

        else:
            raise Warning('Unsupported data type')
            
        
        

                

        
        
        
        ### operations ###

        with self.console.status(f'[gold3]screening {self.type} files', spinner_style = 'white') as status:
            
            
            ## check for duplicate files
            status.start()
            status.update(f'[yellow] checking for duplicate files', spinner_style = 'white')
            self.screen_duplicate_files(fpaths)
            
            if self._duplicates_removed>0:
                self.console.log(f'[yellow]({self.name}) {self._duplicates_removed} duplicate files removed', style = 'notification')
                time.sleep(0.25)
            
            
            ## screen elegible input files
            
            status.update(f'[yellow]screening {len(self._filepaths_in)} {self.type} files', spinner_style = 'white')
            self.__screen_files()
            status.stop()
            status.update(f'', spinner_style = 'white')
            self.console.log(f'[light_green]({self.name}) {self.n_files} {self.type} files available for import [yellow]({len(self._filepaths_failed)} rejected)', style = 'notification')
            time.sleep(.25)
        #self.console.log(f'')
            
            
        self.data_active = self.n_files*[False] # list of indicators for whether a file has been pulled into memory
        self.data = self.n_files*[None]
        #self.labels = [f'file{i}'for i in range(self.n_files)] # label for imported files (search handle)
     
        
    def create_console(self):
        theme = Theme({'screening file': 'gold3 italic',
                       'loading_file': 'dark_cyan',
                       'notification': 'yellow'})
        self.console = Console(tab_size=4,
                               log_path = False,
                               theme = theme) # create a rich-text console 
        
        
    def deactivate_all_data(self):
        self.data = self.n_files*[None]
        self.data_active = self.n_files*[False] 
        
    def screen_duplicate_files(self,fpaths):
        """
        check for duplicate files 

        Parameters
        ----------
        fpaths : list
            list of filepaths to check



        """
        filedata = []# store the filepaths and the size of the main data file for each input file 
        for fpath in fpaths:
            filedata.append([i.split(os.sep)[-1] for i in fpath] + [os.path.getsize(fpath[0])] + [fpath])
        fdata_df = pd.DataFrame(data =filedata, columns = ['fname_main','fname_aux','fsize','fpath'])

        n_files_init= len(fdata_df)
        fdata_df.drop_duplicates(inplace = True, subset = ['fname_main','fname_aux','fsize'])
        

        n_removed = n_files_init - len(fdata_df) # number of files removed
        self._duplicates_removed = n_removed
        self._filepaths_in = [i for i in fdata_df['fpath']]        
        
    # def __Pose_parser(self,Pose):
        
    #     #identify data gaps 
    #     t = p.pose.index.to_numpy() 
    #     dt =np.diff(t, prepend = t[0],).astype(float)/1e9 # time delta in seconds
    #     mask = dt>np.percentile(dt,99) # mask to identify data gaps 

    #     t[mask] = np.datetime64("NaT")
    #     def declump(a):
    #         return [a[s] for s in np.ma.clump_unmasked(np.ma.masked_invalid(a))]
    #     t_seg = declump(t)

    #     x1 = []
    #     x2 = []
    #     for seg in t_seg:
    #         print(len(seg))
    #         x1.append(seg[0])
    #         x2.append(seg[-1])
            
    #     file_start_time = x1
    #     file_end_time = x2
        

    #     return Pose,file_start_time,file_end_time
        
        
    def __pd0_parser(self,fpath,import_file = False,start_frac = 0, end_frac = 1):
        """
        parse a pd0 type file

        Parameters
        ----------
        fpath : TYPE
            DESCRIPTION.
        import_file : TYPE, optional
            DESCRIPTION. The default is False.
        start_frac : TYPE, optional
            DESCRIPTION. The default is 0.
        end_frac : TYPE, optional
            DESCRIPTION. The default is 1.

        Returns
        -------
        data : TYPE
            DESCRIPTION.
        file_start_time : TYPE
            DESCRIPTION.
        file_end_time : TYPE
            DESCRIPTION.

        """
        #screen the file
        

        data  = Pd0(fpath[0], import_file = False)
        file_start_time = pd.to_datetime(data.metadata.first_date_in_file)
        file_end_time = pd.to_datetime(data.metadata.last_date_in_file)


        if import_file:
            start_ensemble = int(start_frac*data.metadata.n_ensembles_in_file)+1
            end_ensemble = int(end_frac*data.metadata.n_ensembles_in_file)
            data = DataSet_pd0(fpath[0], 
                               import_file = True, 
                               start = start_ensemble,
                               end = end_ensemble,
                               pt3_filepath = fpath[1],
                               print_progress = False)
            
     
            
            
        return data, file_start_time,file_end_time
    
    def __ctd_parser(self,fpath,import_file = False,start_time = None, end_time = None):
        """
        parse a ctd type file

        Parameters
        ----------
        fpath : TYPE
            DESCRIPTION.
        import_file : TYPE, optional
            DESCRIPTION. The default is False.
        start_time : TYPE, optional
            DESCRIPTION. The default is none.
        end_time : TYPE, optional
            DESCRIPTION. The default is none.

        Returns
        -------
        data : TYPE
            DESCRIPTION.
        file_start_time : TYPE
            DESCRIPTION.
        file_end_time : TYPE
            DESCRIPTION.

        """
        
        ## add internal seabird method to read from entry a to b !!!!
        # for use in the manager base class
        # global data
        
        data = seabird_ctd(filepath_hex = fpath[0], filepath_xml = fpath[1], import_file = False)
        file_start_time = min(data.data['DateTime'])
        file_end_time = max(data.data['DateTime'])
 
        if import_file:
            data = DataSet_ctd(filepath_hex = fpath[0], filepath_xml = fpath[1], import_file = True)
            
            if type(start_time) == type(None):
                start_time = file_start_time
            
                        
            if type(end_time) == type(None):
                end_time = file_end_time


    
            mask = (data.data['DateTime']>=start_time) & (data.data['DateTime']<=end_time)
            
            #delete data outside of start_frac and end_frac #trim the data 
            #start_idx = max(int(data.n_samples*start_frac),0)
            #end_idx = int(data.n_samples*end_frac)
            
            for key in data.data.keys():
                data.data[key] = data.data[key][mask]#[start_idx:end_idx]
                
            data.n_samples = len(data.data['DateTime'])    
        return data, file_start_time,file_end_time
        
    def __screen_files(self):
        """
        Get metadata from files without reading the entire file. 
        file start and end dates. 

        Parameters
        ----------
        None

        Returns
        -------
        None.

        """
        #screen files initially 
        
        fpaths_to_read = [] # list of filepaths that passes the initial screening 

        for i,fpath in enumerate(self._filepaths_in):
            _pass = True
            reason = 'pass'
            ## check filesize requirements 
            if type(fpath) != type(str()):fsize = round(os.path.getsize(fpath[0])/1000,1)
            else: fsize = round(os.path.getsize(fpath)/1000,1)
            
            if fsize< self._min_fsize:
                _pass = False
                reason = f'file smaller than minimum allowed filesize {self._min_fsize} kb'
            
            ## update failed filepaths and reasons 
            if _pass:
                fpaths_to_read.append(fpath)
            
            else:
                self._filepaths_failed.append(fpath) 
                self._reason_failed.append(reason)
            
        
        for i,fpath in enumerate(fpaths_to_read):
            _pass = True
            
            try:
                self.get_file_metadata(fpath)
                message = f'screening {self.type} files from {self.name} {fpath[0].split(os.sep)[-1]} ({i+1}) '
                #print(message.ljust(1500),end = "\r")
            except Exception as e:
                _pass = False
                reason = e
                
            
            ## update failed filepaths and reasons     
            if not _pass:
                self._reason_failed.append(reason)
                self._filepaths_failed.append(fpath)      
    




                  
        # et = time.time()
        #print(f'\r {150*" "}',end = "\r")
        #print(f'\rCreated dataset "{self.name}" - {self.n_files} {self.type} files ({round(self.dataset_size,1)}kb)')#\n\texecution time: {round(et - st,1)} sec'
        
       
        
    
    
    def get_file_metadata(self,fpath):
        '''
        Get metadata from a file without reading the entire file. 
        
        '''   
        dataset,file_start_time,file_end_time = self._parser(fpath)
        #self.data.append(dataset)
        self.file_start_times.append(file_start_time)
        self.file_end_times.append(file_end_time)
        self.n_files+=1 
        self.filepaths.append(fpath)
        self.dataset_size+=os.path.getsize(fpath[0])
        
        
        
        # check if dataset name is already registered
        proposed_name = dataset.name
        
        i=1
        while proposed_name in self.labels:
            proposed_name = dataset.name + f'_{i}'
            i+=1
            #print('duplicate name')
        self.labels.append(proposed_name)
        
            
        
        
 
        
        
        
        
    # def get_data(label):
    #     retrieve a data

            
            
        
    def get_active_data(self):
        """
        Returns a list of objects for all active files.

        Returns
        -------
        active_data : list
            list of data objects for every active file

        """

        
        active_data = []
        
        for i in range(self.n_files):
            if self.data_active[i]:
                active_data.append(self.data[i])
        
        
        return active_data

    def load_data_between(self,start_time, end_time):
        """
        read data from all files between start_time and end_time. if start_time or 
        end_time fall within a file, then only the portion of the file after 
        start_time and before end_time will be loaded. 

        Parameters
        ----------
        start_time : pd.datetime
            start time to load data from. 
        end_time : TYPE
            end time to load data from. 

        Returns
        -------
        None.

        """
        
        ## filepaths, index of file, start_percentage, end_percentage 
        # global start_pcts, end_pcts,dt



        fnames = []
        indicies = []
        start_pcts = []
        end_pcts = []
        if end_time<=min(self.file_start_times) or start_time>=max(self.file_end_times):
            print('No results in selection')
            #warnings.warn('No results in selection')
        else:    
            
            
            indicies = np.arange(0,self.n_files)
            st= np.array(self.file_start_times)
            et= np.array(self.file_end_times)
            indicies = indicies[((st>=start_time) & (st<=end_time)) | ((et>=start_time) & (et<=end_time)) | ((st<=start_time) & (et>=end_time)) ]
    
    
            
            # indicies = np.arange(start_index,end_index+1)
            fnames = [self.filepaths[i] for i in indicies]
            
            start_pcts = [] # approx percentage into the file where the start time is
            end_pcts   = [] # approx percentage into the file where the end time is
            for idx in indicies:
                st = self.file_start_times[idx]
                et = self.file_end_times[idx]

                dt = et-st
                if et == st:
                    start_pct = 0
                    end_pct   = 1
                else:
                    start_pct = max((start_time - st)/dt,0)
                    end_pct = min((end_time-st)/dt,1)
                    
                start_pcts.append(start_pct)
                end_pcts.append(end_pct)
                    
                    
        
        
        ## load the data   
        #print(f'\r\tloaded ({0}/{len(indicies)}) {self.type} files from {self.name} dataset'.ljust(100), end = '\r')
        
        #print(indicies, start_pcts, end_pcts, start_time, end_time)
        for i,file_index in enumerate(indicies):
            
            args = {'file_index': file_index,'start_frac':start_pcts[i], 'end_frac':end_pcts[i], 'start_time':start_time,'end_time' :end_time} 
            self.load_one_file(args)
            
            
            #print(f'\r ... loading ({i+1}/{len(indicies)}) {self.type} files from {self.name} dataset'.ljust(100), end = '\r')
        #print(f'loaded {i+1} {self.type} files from {self.name} dataset'.ljust(100), end = '\n')    
        #print('\n')     
            # else:
            #     print(f'{fnames[i]} already leaded')
    


    def get_files(self, label=None, index=None):
        """
        Retrieve datasets with matching label or index from the data manager base object.
    
        Parameters
        ----------
        label : scalar, str, or list-like, optional
            Labels of the datasets to load. Defaults to None.
            Examples: '_RDI_005.000', ['_RDI_005.000', '_RDI_005.000_1']
        index : scalar, str, or list-like, optional
            Indices of the datasets to load. Defaults to None.
            Example: 5 or [5]
    
        Returns
        -------
        list
            A list of datasets corresponding to the provided label or index.
    
        Notes
        -----
        If both `label` and `index` are provided, the function will prioritize the index.
        """
        if label and type(label) != list:
            label = [label]
    
        if index and type(index) != list:
            index = [index]
    
        if label and index:
            print('Both index and label supplied - using index')
        elif label:
            index = [item[0] for item in enumerate(self.labels) if item[1] in label]
            
            
    
        if index:
            self.load_files(index)
            datasets = [self.data[i] for i in index]
            
            if len(datasets)==1:
                datasets = datasets[0]
            
            return datasets
        
        else: 
            print(f'No datasets with label {label} or index {index}')
            
    def load_files(self, file_inds):
        """
        Load files using the provided indices and update the relevant data structures.
    
        Parameters
        ----------
        file_inds : list-like
            Indices of the files to load. If not a list, it will be converted to a list.
    
        """
        if type(file_inds) != list:
            file_inds = list(file_inds)
    
        for file_index in file_inds:
            fname = self.filepaths[file_index][0].split(os.sep)[-1]
            if self.type == 'ctd':
                data, file_start_time, file_end_time = self._parser(self.filepaths[file_index], import_file=True)
            else:
                data, file_start_time, file_end_time = self._parser(self.filepaths[file_index], import_file=True)
    
            self.data[file_index] = data
            self.data_active[file_index] = True    
            self.console.log(f'[green4]({self.name}) loaded {fname} ({self.type})')


    def load_one_file(self, args):
        """
        Load data from one file specified by the provided arguments.
    
        Parameters
        ----------
        args : dict
            A dictionary containing:
            file_index : int
                The index of the file to load.
            start_frac : float, optional
                The starting fraction of the file to read (default is None).
            end_frac : float, optional
                The ending fraction of the file to read (default is None).
            start_time : datetime or str, optional
                The starting time from which to begin reading the file (default is None).
            end_time : datetime or str, optional
                The ending time at which to stop reading the file (default is None).
    
        Notes
        -----
        The method checks if the data has already been loaded before attempting to load it.
        It logs the load operation and any failures. If 'ctd' is specified as the type,
        it uses specific time parameters for loading, otherwise, it uses fraction parameters.
        """
        
        file_index = args['file_index']
        start_frac = args.get('start_frac')
        start_time = args.get('start_time')
        end_time = args.get('end_time')
        end_frac = args.get('end_frac')
    
        fname = self.filepaths[file_index][0].split(os.sep)[-1]
        if self.type == 'ctd':
            data, file_start_time, file_end_time = self._parser(self.filepaths[file_index], import_file=True, start_time=start_time, end_time=end_time)
        else:
            data, file_start_time, file_end_time = self._parser(self.filepaths[file_index], import_file=True, start_frac=start_frac, end_frac=end_frac)
    
        self.data[file_index] = data
        self.data_active[file_index] = True
        self.console.log(f'[green4]({self.name}) loaded {fname} ({self.type})')
        # Handle exceptions if necessary

                
    def load_all_data(self, parallel=False):
        """
        Load all registered datasets either sequentially or in parallel.
    
        Parameters
        ----------
        parallel : bool, optional
            If True, loads the datasets in parallel. If False, loads them sequentially.
            Defaults to False.
    
        Notes
        -----
        This function initializes the loading process for all datasets registered in the
        data manager. It can operate in either parallel or sequential mode based on the
        `parallel` flag. For parallel processing, it uses `ptools.parallel_fcn` to manage
        concurrent tasks. Each file's data is loaded using the `load_one_file` method with
        predetermined start and end times, along with start and end fractions.
        """
        if parallel:
            args = [{'file_index': i, 'start_frac': 0, 'end_frac': 1,
                     'start_time': pd.to_datetime('1/1/1899'), 'end_time': pd.to_datetime('1/1/2100')}
                    for i in range(self.n_files)]
            file_data = ptools.parallel_fcn(self.load_one_file, args)
        else:
            for i in range(self.n_files):
                # Uncomment the next line if you want to print progress information
                # print(f'\rloading ({i+1}/{self.n_files}) {self.type} files from {self.name} dataset'.ljust(100), end='\r')
                args = {'file_index': i, 'start_frac': 0, 'end_frac': 1,
                        'start_time': pd.to_datetime('1/1/1899'), 'end_time': pd.to_datetime('1/1/2100')}
                self.load_one_file(args)
    
            

        
    

#%% 

class plotting:
    def __init__(self,dm):
        
        self._dm = dm
        
        # self.dtypes = [i for i in dir(self._dm.data) if not i.endswith('__')]
        

    def timespan(self,interactive = False):  
       
        global legend_items
        # if data == 'adcp':
        #     ds = self._data.adcp
        # if data == 'ctd':
        #     ds = self._data.ctd
        
        
        fig,ax = subplots(figheight = 3,figwidth = 8, for_production = True)
        
        #ax.set_aspect(1/ds.n_files)
 
        
        colors = [dhi_colors.blue2,
                  dhi_colors.green1,
                  dhi_colors.blue3,
                  dhi_colors.green2]
        color = dhi_colors.blue2
        legend_items = []
        for d,dataset in enumerate(self._dm.datasets.keys()):
            ds = self._dm.datasets[dataset]
            x = np.array([ds.file_start_times,ds.file_end_times]).T
            offset = 1
            
            for n in range(ds.n_files):
                if n == 0: label = dataset
                else: label = None
                
                
                #if ds.type == 'ctd':
                gid = '...'+ os.sep + os.sep.join(ds.filepaths[n][0].split(os.sep)[-3:])
  
                
                #print(gid)
                item = ax.plot([x[n,0],x[n,1]],2*[(d)*offset], lw = 20,solid_capstyle='butt', color = color, alpha = 0.7, label = label, gid = gid)

        # Shrink current axis by 20%
        box = ax.get_position()
        ax.set_position([box.x0+.1, box.y0, box.width*0.9, box.height])


        ax.spines['right'].set_visible(False)
        ax.spines['top'].set_visible(False)   
        ax.spines['left'].set_visible(False) 
        
        #ax.get_yaxis().set_visible(False)
        ax.set_ylim(-offset,(d+1)*offset)    
        locator = mpl.dates.AutoDateLocator()
        ax.xaxis.set_major_locator(locator)
        ax.xaxis.set_major_formatter( mpl.dates.AutoDateFormatter(locator) )        
        
        ax.set_xlabel('Time')
    
        ax.set_title(f'Timespan\n{self._dm.name}')
        ax.set_yticks(np.linspace(0,d*offset,d+1))
        ax.minorticks_off()
        

        ax.set_yticklabels([d for d in self._dm.datasets.keys()],fontsize = 10)


        ax.grid(False)
        
        if interactive:
            annot = ax.text(0,0,'TEST')
            def update_annot(text,x,y):
                annot.set_visible(True)
                annot.set_position((x,y))
                annot.set_text(text)
                fig.canvas.draw()
                
            def on_plot_hover(event):
                global curve
                
                
                annot.set_visible(False)
                for curve in ax.get_lines():
                    if curve.contains(event)[0]:
                        x, y = event.xdata, event.ydata
                        text = curve.get_gid()
                        annot.set_visible(True)
                        annot.set_position((x,y))
                        annot.set_text(text)
                        fig.canvas.draw()
                        curve.set_color(dhi_colors.red2)
                    else:
                        curve.set_color(dhi_colors.blue2)
                    
            fig.canvas.mpl_connect('motion_notify_event', on_plot_hover)           
            plt.show()
        
        
        return fig, ax
                
    
#%%

class DataManager:
    def __init__(self, name='Data Manager'):
        """
        Initialize the DataManager with a given name and set up its environment.
        
        The DataManager class is designed to manage and handle many DataSet objects, particularly focusing on operations such as loading, unloading, and serializing data.

        Parameters
        ----------
        name : str, optional
            The name assigned to the DataManager instance. Defaults to 'Data Manager'.

        """
        self.name = name
        self.datasets = {}
        self.n_datasets = 0
        self.plot = plotting(self)
        
        self.create_console()

        #
        text = Text(f'Data Manager ({name})')

        self.console.print(f'Data Manager ({name})', style = 'header')
        
        
        
    def create_console(self):
        """
        Create and configure a rich console for the data manager.
    
        Notes
        -----
        This method initializes a rich console object with a specific theme and configurations.
        It is often necessary to invoke this method after loading a data manager object from a
        serialized state to re-establish the console functionalities that may not persist through
        serialization. The console uses a defined theme for headers and new dataset notifications
        and is configured not to log to a path while using a standard color system.
    
        Examples
        --------
        >>> dm.create_console()
        """
        theme = Theme({'header': 'bold cyan underline',
                       'new_dataset': 'dark_cyan'})
        self.console = Console(tab_size=4,
                               theme=theme,
                               log_path=False,
                               color_system='standard')

            
    def deactivate_all_data(self):
        """
        Deactivate all active data across all datasets managed by the data manager.
    
        Notes
        -----
        This method iterates through each dataset registered in the data manager and calls
        the `deactivate_all_data` method for each dataset. This is typically used to free up memory
        by unloading data that is currently loaded but no longer needed.
        """
        for key in self.datasets.keys():
            self.datasets[key].deactivate_all_data()

            
    def add_dataset(self, file_type, fpaths, name, min_fsize=0):
        """
        Register a new dataset with the data manager based on the specified file type and paths.
    
        Parameters
        ----------
        file_type : str
            Type of the file associated with the dataset, typically 'pd0' or 'ctd'.
        fpaths : tuple
            A tuple of file paths, specific to the file type:
            - For 'pd0': (<pd0 filepath>, <pt3 filepath>)
            - For 'ctd': (<hex filepath>, <xmlcon filepath>)
        name : str
            The name of the dataset to be registered under the data manager.
        min_fsize : int, optional
            The minimum file size in bytes to consider for inclusion in the dataset. Defaults to 0.
    
        Notes
        -----
        This method adds a collection of a certain dataset type to the data manager's dataset registry.
        It filters out files below the specified minimum file size, checks for duplicates,
        and verifies that they can be read properly. Once a dataset is successfully registered,
        it logs this event to the console.
    
        Examples
        --------
        >>> dm.add_dataset('pd0', (r'/path/to/pd0file.pd0', r'/path/to/pt3file.pt3'), 'ocean_currents', 1000)
        """
        self.console.log(f'({name}) dataset registered', style='new_dataset')
        time.sleep(0.25)
    
        if file_type == 'pd0':
            self.datasets[name] = manager_base(parser_name='pd0', name=name, fpaths=fpaths, min_fsize=min_fsize)
        elif file_type == 'ctd':
            self.datasets[name] = manager_base(parser_name='ctd', name=name, fpaths=fpaths, min_fsize=min_fsize)
        self.n_datasets += 1
    
            

        

        
    
        
    def load_all_data_between(self, start_time, end_time):
        """
        Load data from all registered datasets between specified start and end times.
    
        Parameters
        ----------
        start_time : pd.Timestamp
            The start time from which to begin loading data.
        end_time : pd.Timestamp
            The end time until which data should be loaded.
    
        Returns
        -------
        None
    
        Notes
        -----
        This method iterates over all datasets managed by the data manager and invokes
        `load_data_between` on each, loading data that falls within the specified time range.
        This is particularly useful for temporal data analyses where data needs to be synchronized
        across multiple datasets.
    
        See Also
        --------
        data_manager.load_data_between : Load data for a single dataset between two time points.
        """
        # Uncomment the next line if you want to provide real-time status updates using the console
        # with self.console.status(f'[yellow] Loading all data between {start_time.strftime("%d%b%y %H:%M")} and {end_time.strftime("%d%b%y %H:%M")}', spinner_style='white') as status:
        for key in self.datasets.keys():
            self.datasets[key].load_data_between(start_time=start_time, end_time=end_time)

                
    def load_all_data(self):
        """
        Load all data for each dataset managed by the data manager.
    
        Notes
        -----
        This method iterates over all datasets registered in the data manager and calls
        their respective `load_all_data` method to load their data. This function is useful
        for initializing or reloading all datasets at once, ensuring that they are ready for processing.
        """
        # Uncomment the next line if you want to provide real-time status updates using the console
        # with self.console.status(f'[yellow] Loading all data', spinner_style='white') as status:
        for key in self.datasets.keys():
            self.datasets[key].load_all_data()

            
    def save(self, fpath=None):
        """
        Serialize the data manager object and save it to a file.
    
        Parameters
        ----------
        fpath : str, optional
            The file path where the data manager object will be saved.
            If not specified, it defaults to a file in the current directory named after the data manager's name with a '.dm' extension.
            If provided, '.dm' will be appended to the specified path.
    
        Notes
        -----
        This method serializes the entire data manager object into a binary format using dill.
        Before serialization, it deactivates all loaded data and sets the console attributes to None to prevent issues during serialization.
        After saving, it reconstructs the console objects for the data manager and its datasets.
        """
        if fpath is None:
            fpath = '.' + os.sep + self.name + '.dm'
        else:
            fpath = fpath + '.dm'
    
        with open(fpath, "wb") as dill_file:
            # Deactivate all loaded data and disable print consoles for serialization
            self.deactivate_all_data()
            self.console = None
            for dataset in self.datasets:
                self.datasets[dataset].console = None
    
            # Serialize the data manager object and save to file
            dill.dump(self, dill_file)
    
            # Recreate consoles
            self.create_console()
            for dataset in self.datasets:
                self.datasets[dataset].create_console()

            
            
    def find_pairs(self, dataset1_name='ADCP', dataset2_name='CTD', get='label', min_overlap='5min'):
        """
        Identify pairs of datasets between two specified types where there is a significant time overlap.
    
        Parameters
        ----------
        dataset1_name : str, optional
            The name of the first dataset type. Defaults to 'ADCP'.
        dataset2_name : str, optional
            The name of the second dataset type. Defaults to 'CTD'.
        get : str, optional
            Specifies the type of return value: 'label' for dataset labels, 'index' for dataset indices.
            Defaults to 'label'.
        min_overlap : str or pandas.Timedelta, optional
            The minimum duration of overlap required to consider two datasets as a pair.
            Can be specified as a string that pandas.Timedelta can parse. Defaults to '5min'.
    
        Returns
        -------
        list of tuples
            A list of tuples representing either the labels or indices of the dataset pairs that
            have a time overlap greater than the specified minimum.
    
        Notes
        -----
        The function calculates time overlaps between files in the specified datasets.
        If the overlap exceeds the specified minimum threshold, the pair is added to the results.
        This function utilizes `ptools.calculate_time_overlap` to calculate the actual time overlaps
        between two time intervals. If the `get` parameter is 'label', it returns labels, otherwise, it returns indices.
        """
        ds1_start_times = self.datasets[dataset1_name].file_start_times
        ds1_end_times = self.datasets[dataset1_name].file_end_times
        n_ds1s = self.datasets[dataset1_name].n_files
    
        ds2_start_times = self.datasets[dataset2_name].file_start_times
        ds2_end_times = self.datasets[dataset2_name].file_end_times
        n_ds2s = self.datasets[dataset2_name].n_files
    
        min_overlap = pd.to_timedelta(min_overlap)
    
        pairs = []
        for i in range(n_ds1s):
            ds1_st = ds1_start_times[i]
            ds1_et = ds1_end_times[i]
    
            for j in range(n_ds2s):
                ds2_st = ds2_start_times[j]
                ds2_et = ds2_end_times[j]
    
                # Calculate time overlap
                t_overlap = ptools.calculate_time_overlap(ds1_st, ds1_et, ds2_st, ds2_et)
                if t_overlap > min_overlap:
                    ds1_label = self.datasets[dataset1_name].labels[i]
                    ds2_label = self.datasets[dataset2_name].labels[j]
    
                    ds1_index = i
                    ds2_index = j
    
                    if get == 'label':
                        pairs.append((ds1_label, ds2_label))
                    else:
                        pairs.append((ds1_index, ds2_index))
    
        return pairs

        
        
        

        

#%%   load save and merge functions (serialization)
import copy


def concatenate(dms,name):
    # concatenate a list of DataManager objects. 
    # new DataManager will be given name (name)
    dm = DataManager(name = name)    
    for dm_c in dms:
        # get all datasets in each data manager object to combine 
        for name,dataset in dm_c.datasets.items():
            dataset.console = None
            dm.datasets[name] = copy.deepcopy(dataset)
            dataset.create_console()
        dm.n_datasets = len(dm.datasets)
    return dm
    

        
        
        
def load(fpath):
    # deserialize a data manager object 
    with open(fpath, "rb") as dill_file:
        # deserialize data manager object and reinitilize the print consoles
         dm = dill.load(dill_file)    
         dm.create_console()
         for dataset in dm.datasets:
             dm.datasets[dataset].create_console()
             
    return dm






import numpy as np
import os 
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.dates as mdates
from matplotlib.collections import LineCollection
import matplotlib.ticker as mticker
from matplotlib.ticker import AutoLocator
import sys
import seaborn as sns
import pandas as pd
import cmocean
# #if __name__ == '__main__':
# sys.path.insert(1, r'C:\Users\anba\Desktop\Projects\pyplume\pyplume\ptools')
# import ptools

# sys.path.insert(1, r'C:\Users\anba\Desktop\Projects\pyplume\pyplume\plotting')
# from matplotlib_shell import subplots
    
    
# else:
from ..ptools import ptools
from ..plotting.matplotlib_shell import subplots, dhi_colors



#from ..plotting import plotting





class plotting:
    def __init__(self,pd0):
        
        self._pd0 = pd0
        
        
        
        self._cmaps =  {'PERCENT GOOD': plt.cm.binary,
                     'ECHO INTENSITY': plt.cm.turbo,
                     'FILTERED ECHO INTENSITY': plt.cm.turbo,
                     'CORRELATION MAGNITUDE': plt.cm.nipy_spectral,
                     'ABSOLUTE BACKSCATTER':plt.cm.turbo,
                     'NTU':plt.cm.turbo,
                     'SSC':plt.cm.turbo,
                     'SIGNAL TO NOISE RATIO':plt.cm.bone_r}


    def four_beam_flood_plot(self,**kwargs):#plot_by = 'bin',start_bin = None,end_bin = None,start_ensemble = None,end_ensemble = None,title = None):
        """
        Generate a flooded color plot
    
        Parameters
        ----------
        self._pd0 : object
            WorkhorseADCP object.
        plot_by : str
            y-axes plot method ('bin','depth').
        start_bin : int
            First bin to plot. (use zero based index)
        end_bin : int
            Last bin to plot.(use zero based index)
        start_ensemble : int
            First ensemble to plot.(use zero based index)
        end_ensemble : int
            Last ensemble to plot.(use zero based index)
        title : str
            plot axes title.
        field_name: str
            ensemble data type (string). "ECHO INTENSITY" (default), "CORRELATION MAGNITUDE", "PERCENT GOOD", others in ensemble_fields
    
        Returns
        -------
        fig,ax
            matplotlib figure and axes objects
    
        """
        
        # if self._pd0.ensemble_data[0]['COORDINATE SYSTEM'] != 'BEAM COORDINATES':
        #     print('Plot feature only implemented for BEAM coordinates')
    
    
        plot_by = kwargs.pop('plot_by','bin')  
        start_bin = kwargs.pop('start_bin',0)
        end_bin = kwargs.pop('end_bin',self._pd0.config.number_of_cells-1)    
        start_ensemble = kwargs.pop('start_ensemble',0)
        end_ensemble = kwargs.pop('end_ensemble',self._pd0.n_ensembles)       
        title = kwargs.pop('title',self._pd0.metadata.filepath.split(os.sep)[-1])    
        field_name = kwargs.pop('field_name','ECHO INTENSITY')
        ctd = kwargs.pop('ctd',None)
        
        extra_rows = kwargs.pop('extra_rows',0)
        
        nbins = (end_bin - start_bin)
        X  = self._pd0.get_ensemble_array(field_name = field_name)[:,start_bin:end_bin,start_ensemble:end_ensemble]
        
        x1 = X[0]
        x2 = X[1]
        x3 = X[2]
        x4 = X[3]

        #echo_intensity = self._pd0.get_ensemble_array(beam_number = 0, field_name = 'CORRELATION MAGNITUDE')[start_bin:end_bin,start_ensemble:end_ensemble]
        ensemble_times = self._pd0.get_ensemble_datetimes()[start_ensemble:end_ensemble]
        
        subplot_titles = []
        nrow = self._pd0.config.number_of_beams
        if ctd:
            fig,ax = subplots(nrow = nrow+1 + extra_rows, ncol = 1, figheight = 8, figwidth = 10.5, width_ratios = [1], sharex = True, sharey = False)
            # Manually link the y-axes of the first four subplots
            base_ax = ax[0]  # Use the first subplot as the base for sharing
            for axes in ax[1:nrow]:  # Link axes 2, 3, 4 to the first
                axes.sharey(base_ax)
            
            # Optionally, you can use the below loop to enforce the linked behavior visually
            # It syncs all linked axes limits after plotting
            #base_ax.get_shared_y_axes().join(*ax[:nrow])
            base_ax.autoscale()
        else:
            fig,ax = subplots(nrow = nrow + extra_rows, ncol = 1, figheight = 8, figwidth = 10.5, width_ratios = [1], sharex = True, sharey = True)
    
        ## format the ADCP data axes (left)
        topax = ax[0].twiny()
        topax.set_xlim(start_ensemble,end_ensemble)
        ax[0].set_zorder(4)
        #topax.set_xticks(np.arange())
        ax[0].set_title(field_name)
        
        fig.suptitle(title, fontsize = 16)
        

            # for i,axes in enumerate(ax[:self._pd0.config.number_of_beams]):
            #     axes.get_shared_y_axes().remove(ax[-1])
            #     ax[-1].get_shared_y_axes().remove(axes)
    
        # set plot params based on instrument configuration 
    
        def set_axes_labels(axs,ylabel):
            # function to set the ylabel on a list of axes objects 
            for ax in axs:
                ax.set_ylabel(ylabel, fontsize = 8)
                
            return axs
                
        
        
        # set plotting extents in vertical direction
        if plot_by == 'bin':
            ylims = [start_bin,end_bin]
            set_axes_labels(ax,'Bin')
        elif plot_by == 'depth':
            bin_depths = self._pd0.geometry.get_bin_midpoints_depth()
            ylims = [bin_depths[start_bin],bin_depths[end_bin]]
            set_axes_labels(ax,'Depth')
        elif plot_by == 'HAB':
            bin_heights = self._pd0.geometry.get_bin_midpoints_HAB()
            ylims = [bin_heights[start_bin],bin_heights[end_bin]]
            set_axes_labels(ax,'Height Above Bed (m)')
            #fig.text(0.02,0.5, 'Height Above Bed (m)', va='center', rotation='vertical',fontsize = 12)
        else: 
            print('Invalid plot_by parameter')
            
            
            
        
        
        def get_plot_extent(X):
            """
            set the image plot extent based on the ADCP beam configuration
    
            Parameters
            ----------
            X : numpy array
                Input array (ensemble data).
    
            Returns
            -------
            X : numpy array
                Possibly rotated input array (ensemble data) - to match the sensor orientation.
                
            extent : list
                bounding box for the plotted image.
    
            """
       
        
            # set plotting extents in horizontal direction
            xlims = mdates.date2num(ensemble_times) # list of elegible xlimits 
            if self._pd0.config.beam_facing == 'UP':
                
                extent = [xlims[0],xlims[-1],ylims[0],ylims[1]]
            else:
                X = np.flipud(X)
                extent = [xlims[0],xlims[-1],ylims[1],ylims[0]]
            return X,extent
            
        
    
        cmap = self._cmaps.pop(field_name,'turbo')
        
        dataset = [x1,x2,x3,x4] # hold all elegible beam datasets 
        dataset = dataset[:self._pd0.config.number_of_beams] # just plot for valid beams
        
        for x in dataset:
            x[x == 32768] = np.nan
        
        
        
        if field_name == 'PERCENT GOOD':
            vmin = 0
            vmax = 100
        elif field_name == 'SIGNAL TO NOISE RATIO':
            vmin = np.nanmin(dataset)
            vmax = 50
        elif field_name == 'ABSOLUTE BACKSCATTER':
            vmin = -95
            vmax = -10
        else:
            vmin = np.nanmin(dataset)
            vmax = np.nanmax(dataset)
        
        ## overwrite with kwargs if provided 
        vmin = kwargs.pop('vmin',vmin) 
        vmax = kwargs.pop('vmax',vmax) 
        
        
        for i,x in enumerate([x1,x2,x3,x4]):
            xi,extent = get_plot_extent(x)
            if self._pd0.config.beam_facing == 'UP': 
                origin = 'lower'
            
            else: origin = 'upper'
            
            im = ax[i].matshow(x, origin = origin, extent = extent, cmap = cmap, aspect = 'auto', resample = False, vmin = vmin, vmax = vmax)
            
        
        
        for i,axes in enumerate(ax[:self._pd0.config.number_of_beams]):# modify subplot apperance 
            # set colorbar 
            
            
            #if i==(len(ax)-1) and ctd: break # dont plot a colorbar if 
            cbar = fig.colorbar(im, ax=axes,orientation="vertical",location = 'right',fraction=0.046)
            cbar.set_label(f'Beam {i+1}', rotation=90,fontsize= 8)
            
            axes.xaxis.set_major_locator(mticker.FixedLocator(axes.get_xticks()))
            axes.xaxis.set_major_formatter(mdates.DateFormatter('%H:%M %d%b%y '))
            axes.grid(alpha = 0.1)
            axes.set_xticklabels(axes.get_xticklabels(), rotation = 0, ha = 'center')
            axes.xaxis.tick_bottom()
            axes.xaxis.set_label_position('bottom') 
            
            axes.yaxis.set_major_locator(AutoLocator())

                        
            
        if ctd:
    
            #ax[-1].autoscale(enable=True, axis='y', tight=False)
            # group = ax[-1].get_shared_y_axes()
            # group.remove(ax[-1])
            df = ctd.to_pandas_df()
            df = df.reindex(ensemble_times, method = 'nearest', tolerance = pd.to_timedelta('60s')).dropna()

            
            
            if field_name == 'SSC':
                #t,ssc = ctd.get_timeseries_data(field_name = 'SSC (mg/L)')
                t = df.index
                ssc = df['SSC (mg/L)']
                sc = ax[-1].scatter(t,ssc,c= ssc,vmax = vmax,vmin = vmin, alpha = 0.9, label = 'SSC (CTD)')
                ax[-1].set_ylabel('SSC (mg/L)', fontsize = 12)
                
                cbar = fig.colorbar(im, ax=ax[-1],orientation="vertical",location = 'right',fraction=0.046)
                cbar.set_label(f'CTD SSC (mg/L)', rotation=90,fontsize= 8)
            else:
                
                #t,turb = ctd.get_timeseries_data(field_name = 'Turbidity (NTU)')
                
                t = df.index
                turb = df['Turbidity (NTU)'].to_numpy()
                sc = ax[-1].scatter(t,turb,c= 'black', alpha = 0.9, s = 0.25, label = 'Turbidity (CTD)')
                ax[-1].plot(t,turb,c= 'black', alpha = 0.9,lw = 1, label = 'Turbidity (CTD)')
                ax[-1].set_ylabel('Turbidity (NTU)', fontsize = 8)
                ax[-1].set_ylim(0,1.1*np.nanmax(turb))
                
                ax[-1].yaxis.set_major_locator(AutoLocator())
                
                cbar = fig.colorbar(im, ax=ax[-1],orientation="vertical",location = 'right',fraction=0.046)
                cbar.set_label(f'CTD Turbidity (NTU)', rotation=90,fontsize= 8)
            #     ax[i].scatter(df.index,df.z,c =df['Turbidity (NTU)'], s = 15, edgecolor = None, cmap = cmap, marker = 'o', alpha = 0.9, label = 'Turbidity (CTD)')
            #ax[-1].xaxis.set_major_locator(AutoLocator())
                        
            locator = mdates.AutoDateLocator(minticks=3, maxticks=7)
            formatter = mdates.ConciseDateFormatter(locator)
            formatter.formats = ['%D',  # ticks are mostly years
                     '%b%y',       # ticks are mostly months
                     '%b%d',       # ticks are mostly days
                     '%H:%M',    # hrs
                     '%H:%M',    # min
                     '%S.%f', ]  # secs
            ax[-1].xaxis.set_major_locator(locator)
            ax[-1].xaxis.set_major_formatter(formatter)
            #ax[-1].xaxis.set_major_locator(mticker.FixedLocator(ax[-1].get_xticks()))
            #ax[-1].xaxis.set_major_formatter(mdates.DateFormatter('%H:%M %d%b%y '))
            ax[-1].grid(alpha = 0.1)
            #ax[-1].set_xticklabels(ax[-1].get_xticklabels(), rotation = 0, ha = 'center')
            ax[-1].xaxis.tick_bottom()
            ax[-1].xaxis.set_label_position('bottom') 

        
        return fig,ax  

        
    
    def four_beam_mesh_plot(self,**kwargs):#plot_by = 'bin',start_bin = None,end_bin = None,start_ensemble = None,end_ensemble = None,title = None):
        """
        Generate a flooded color plots of correlation magnitude
    
        Parameters
        ----------
        self._pd0 : object
            WorkhorseADCP object.
        plot_by : str
            y-axes plot method ('bin','depth').
        start_bin : int
            First bin to plot. (use zero based index)
        end_bin : int
            Last bin to plot.(use zero based index)
        start_ensemble : int
            First ensemble to plot.(use zero based index)
        end_ensemble : int
            Last ensemble to plot.(use zero based index)
        title : str
            plot axes title.
        field_name: str
            ensemble data type (string). "ECHO INTENSITY" (default), "CORRELATION MAGNITUDE", "PERCENT GOOD", others in ensemble_fields
    
        Returns
        -------
        fig,ax
            matplotlib figure and axes objects
    
        """
        
        # if self._pd0.ensemble_data[0]['COORDINATE SYSTEM'] != 'BEAM COORDINATES':
        #     print('Plot feature only implemented for BEAM coordinates')
    
        
        plot_by = kwargs.pop('plot_by','bin')  
        start_bin = kwargs.pop('start_bin',0)
        end_bin = kwargs.pop('end_bin',self._pd0.config.number_of_cells)    
        start_ensemble = kwargs.pop('start_ensemble',0)
        end_ensemble = kwargs.pop('end_ensemble',self._pd0.n_ensembles)       
        title = kwargs.pop('title',self._pd0.metadata.filepath.split(os.sep)[-1])    
        field_name = kwargs.pop('field_name','ECHO INTENSITY')
        ctd = kwargs.pop('ctd',None)
        
        extra_rows = kwargs.pop('extra_rows',0)
        
        nbins = (end_bin - start_bin)
        X  = self._pd0.get_ensemble_array(field_name = field_name)[:,start_bin:end_bin,start_ensemble:end_ensemble]
        
        x1 = X[0]
        x2 = X[1]
        x3 = X[2]
        x4 = X[3]
    
        #echo_intensity = self._pd0.get_ensemble_array(beam_number = 0, field_name = 'CORRELATION MAGNITUDE')[start_bin:end_bin,start_ensemble:end_ensemble]
        ensemble_times = self._pd0.get_ensemble_datetimes()[start_ensemble:end_ensemble]
        
        subplot_titles = []
        nrow = 4 #self._pd0.config.number_of_beams 
    
        fig,ax = subplots(nrow = nrow + extra_rows, ncol = 1, figheight = 8, figwidth = 10.5, width_ratios = [1], sharex = True, sharey = True)
    
        ## format the ADCP data axes (left)
        # topax = ax[0].twiny()
        # topax.set_xlim(start_ensemble,end_ensemble)
        # ax[0].set_zorder(4)
        #topax.set_xticks(np.arange())
        ax[0].set_title(field_name)
        
        fig.suptitle(title, fontsize = 16)
        
    
        # set plot params based on instrument configuration 
    
        def set_axes_labels(axs,ylabel):
            # function to set the ylabel on a list of axes objects 
            for ax in axs:
                ax.set_ylabel(ylabel)
                
            return axs
        
        # set plotting extents in vertical direction
        if plot_by == 'bin':
           # ylims = [start_bin,end_bin]
            set_axes_labels(ax,'Bin Dist. (m)')
            Z = self._pd0.geometry.get_relative_beam_midpoint_positions(mask = False)[2] # z mesh values
            
            z_platform =  (self._pd0.config.bin_1_distance/100) +np.zeros(self._pd0.geometry.get_corrected_bottom_track().shape)
            
            print(z_platform.min())
            ylims = [np.nanmin(Z),np.nanmax(Z)]
            
        elif plot_by == 'depth':
            #bin_depths = self._pd0.get_bin_midpoints_depth()
            #ylims = [bin_depths[start_bin],bin_depths[end_bin]]
            set_axes_labels(ax,'Depth (m)')
            Z = self._pd0.geometry.get_absolute_beam_midpoint_positions(mask = False)[2] # z mesh values
            
            z_platform = np.repeat(self._pd0.geometry.pose.pose['z'].to_numpy()[np.newaxis,:],repeats = self._pd0.config.number_of_beams,axis = 0)
            ylims = [np.nanmin(Z),np.nanmax(Z)]
            
            
            
        elif plot_by == 'HAB':
            #bin_heights = self._pd0.get_bin_midpoints_HAB()
            #ylims = [bin_heights[start_bin],bin_heights[end_bin]]
            set_axes_labels(ax,'HAB (m)')
            Z = self._pd0.geometry.get_absolute_beam_midpoint_positions_HAB(mask = False)[2] # z mesh values
            z_platform = -  self._pd0.geometry.get_corrected_bottom_track()
            
            ylims = [0,1.1*np.nanmax(z_platform[:,start_ensemble:end_ensemble])]
        else: 
            print('Invalid plot_by parameter')
          
        
        z_platform = z_platform[:,start_ensemble:end_ensemble]
        Z = Z[:,start_bin:end_bin,start_ensemble:end_ensemble]
    
        cmap = kwargs.pop('cmap',self._cmaps.pop(field_name,'turbo'))
        
        dataset = [x1,x2,x3,x4] # hold all elegible beam datasets 
        dataset = dataset[:self._pd0.config.number_of_beams] # just plot for valid beams
        
        for x in dataset:
            x[x == 32768] = np.nan
        if field_name == 'PERCENT GOOD':
            vmin = 0
            vmax = 100
            print('pg')
        elif field_name == 'SIGNAL TO NOISE RATIO':
            vmin = np.nanmin(dataset)
            vmax = 50
            print('stn')
        elif field_name == 'ABSOLUTE BACKSCATTER':
            vmin = -95
            vmax = -45
        elif field_name == 'NTU':
            vmin = np.nanmin(dataset)
            vmax = np.nanmax(dataset)
        else:
            vmin = np.nanmin(dataset)
            vmax = np.nanmax(dataset)
        
        ## overwrite with kwargs if provided 
        vmin = kwargs.pop('vmin',vmin) 
        vmax = kwargs.pop('vmax',vmax) 
        
        
        for i,x in enumerate([x1,x2,x3,x4]):
        #     xi,extent = get_plot_extent(x)
        #     if self._pd0.config.beam_facing == 'UP': 
        #         origin = 'lower'
        #     else: origin = 'upper'
            
            #im = ax[i].matshow(x, origin = origin, extent = extent, cmap = cmap, aspect = 'auto', resample = False, vmin = vmin, vmax = vmax)
            
            
            
            #T = np.repeat(self._pd0.get_ensemble_datetimes(),self._pd0.config.number_of_cells).reshape((len(self._pd0.get_ensemble_datetimes()),self._pd0.config.number_of_cells)).T
            T = np.repeat(ensemble_times,self._pd0.config.number_of_cells).reshape((len(ensemble_times),self._pd0.config.number_of_cells)).T[start_bin:end_bin]#,start_ensemble:end_ensemble]
            #Z = X_hab[2,beam_no]
            #print(Z[i].shape,T.shape,x.shape,z_platform[i].shape)
            im = ax[i].pcolor(T,Z[i],x, vmin = vmin, vmax = vmax, cmap = cmap)#,antialiaseds = True)
            
            
            ax[i].plot(ensemble_times,z_platform[i], label = 'ROV (z)', lw = 1, alpha = 0.7)
    
            
            #axp.set_ylabel('Turbidity')
        
        for i,axes in enumerate(ax[:self._pd0.config.number_of_beams]):# modify subplot apperance 
            # set colorbar 
            
            
            #if i==(len(ax)-1) and ctd: break # dont plot a colorbar if 
        
            
            cbar = fig.colorbar(im, ax=axes,orientation="vertical",location = 'right',fraction=0.046, extend = 'max')
            cbar.set_label(f'Beam {i+1}', rotation=90,fontsize= 10)
            axes.grid(alpha = 0.1)
            axes.xaxis.tick_bottom()
            axes.xaxis.set_label_position('bottom') 
            axes.set_ylim(ylims)
            
            
            
        ax[-1].set_xlabel('Time')
    
       
        if ctd:
            for i,axes in enumerate(ax[:self._pd0.config.number_of_beams]):
                # t,turb = ctd.get_timeseries_data(field_name = 'Turbidity (NTU)')
                

                
                df = ctd.to_pandas_df()
                df = df.reindex(ensemble_times, method = 'nearest')
                df['z'] = z_platform[i]+0.25
                #df = df.resample('1s').mean()
                
                if field_name == 'NTU':
                    axes.scatter(df.index,df.z,c =df['Turbidity (NTU)'], vmin = vmin,vmax = vmax, s = 10, edgecolor = None, cmap = cmap, marker = 's' , alpha = 0.9)#, label = 'Turbidity (CTD)')
                    #axp.set_ylabel('Turbidity (NTU)', fontsize = 7)
                elif field_name == 'SSC':
                    axes.scatter(df.index,df.z,c =df['SSC (mg/L)'], vmin = vmin,vmax = vmax, s = 10, edgecolor = None, cmap = cmap, marker = 's' , alpha = 0.9)#, label = 'SSC (CTD)')
                   # axp.set_ylabel('SSC (mg/L)', fontsize = 7)
                else:
                    axp = ax[i].twinx()
                    axp.plot(df.index,df['Turbidity (NTU)'], lw = 0.25, c = 'black', alpha = 0.6)
                    axp.set_zorder(4)
                    axp.set_ylabel('Turbidity (NTU)', fontsize = 7)
                    
                    
                    
                # else:
                #     ax[i].scatter(df.index,df.z,c =df['Turbidity (NTU)'], s = 15, edgecolor = None, cmap = cmap, marker = 'o', alpha = 0.9, label = 'Turbidity (CTD)')
                    
        for i,axes in enumerate(ax):
            axes.legend(fontsize = 7)
            
        if extra_rows>0:    
            for axes in ax[-1:]  :
                cbar = fig.colorbar(im, ax=axes,orientation="vertical",location = 'right',fraction=0.046, extend = 'max')
                #cbar.set_label( rotation=90,fontsize= 8)
                axes.grid(alpha = 0.1)
                axes.xaxis.tick_bottom()
                axes.xaxis.set_label_position('bottom') 
                axes.set_ylim(ylims)
            
        # unlink the axisd
        # if extra_rows>0:
        #     for i in range(1,extra_rows+1):
        #         axes.get_shared_y_axes().remove(ax[-i])
        #         ax[-i].get_shared_y_axes().remove(axes)
        
        return fig,ax   
         
    def single_beam_flood_plot(self,**kwargs):#plot_by = 'bin',start_bin = None,end_bin = None,start_ensemble = None,end_ensemble = None,title = None):
        """
        Generate a fencegate plot of ensemble data
        
        Parameters
        ----------
        self._pd0 : object
            WorkhorseADCP object.
        plot_by : str
            y-axes plot method ('bin','depth').
        start_bin : int
            First bin to plot. (use zero based index)
        end_bin : int
            Last bin to plot.(use zero based index)
        start_ensemble : int
            First ensemble to plot.(use zero based index)
        end_ensemble : int
            Last ensemble to plot.(use zero based index)
        title : str
            plot axes title.
        field_name: str
            ensemble data type (string). "ECHO INTENSITY" (default), "CORRELATION MAGNITUDE", "PERCENT GOOD", others in ensemble_fields
        beam : int or str. 
            beam number to plot. if 0 or 'average', the beam average is returned. If none, the beam average is returned. 
        
        Returns
        -------
        fig,ax
            matplotlib figure and axes objects
        
        """
        
        plot_by = kwargs.pop('plot_by','bin')  
        start_bin = kwargs.pop('start_bin',0)
        end_bin = kwargs.pop('end_bin',self._pd0.config.number_of_cells)    
        start_ensemble = kwargs.pop('start_ensemble',0)
        end_ensemble = kwargs.pop('end_ensemble',self._pd0.n_ensembles)       
        title = kwargs.pop('title',self._pd0.metadata.filepath.split(os.sep)[-1])    
        field_name = kwargs.pop('field_name','ECHO INTENSITY')
        beam = kwargs.pop('beam',0)
        if beam == 'average': beam = 0 
        
        
        nbins = (end_bin - start_bin)
        x = self._pd0.get_ensemble_array(field_name = field_name)[start_bin:end_bin,start_ensemble:end_ensemble]
        
        if beam == 0:
            x = np.nanmean(x,axis = 0)
        else:
            x = x[beam-1,:,:]
        
        #echo_intensity = self._pd0.get_ensemble_array(beam_number = 0, field_name = 'CORRELATION MAGNITUDE')[start_bin:end_bin,start_ensemble:end_ensemble]
        ensemble_times = self._pd0.get_ensemble_datetimes()[start_ensemble:end_ensemble]
        
        
        subplot_titles = []
        fig,ax = subplots( figheight = 2.5, figwidth = 10.5, width_ratios = [1])
        
        
        ## format the ADCP data axes (left)
        topax = ax.twiny()
        ax.set_title(title)
        
        
        # set plotting extents in vertical direction
        if plot_by == 'bin':
            ylims = [start_bin,end_bin]
            ax.set_ylabel('Bin')
        elif plot_by == 'depth':
            bin_depths = self._pd0.get_bin_midpoints_depth()
            ylims = [bin_depths[start_bin],bin_depths[end_bin]]
            ax.set_ylabel('Depth')
        elif plot_by == 'HAB':
            bin_heights = self._pd0.get_bin_midpoints_HAB()
            ylims = [bin_heights[start_bin],bin_heights[end_bin]]
            ax.set_ylabel('Height Above Bed')
        else: 
            print('Invalid plot_by parameter')
        
        
        # set plotting extents in horizontal direction
        xlims = mdates.date2num(ensemble_times) # list of elegible xlimits 
        if self._pd0.config.beam_facing == 'UP':
            extent = [xlims[0],xlims[-1],ylims[0],ylims[1]]
        else:
            x = np.flipud(x)
            extent = [xlims[0],xlims[-1],ylims[1],ylims[0]]
         
        cmap = self._cmaps[field_name]
        
        
        
        im = ax.matshow(x, origin = 'lower', extent = extent, cmap = cmap, aspect = 'auto',resample = False )
        cbar = fig.colorbar(im, ax=ax,orientation="vertical",location = 'right',fraction=0.046)
        
        if beam == 0:
            cbar_label = f'{field_name}\n BEAM AVERAGE' 
        else:
            cbar_label = f'{field_name}\n BEAM {beam}' 
        cbar.set_label(cbar_label, rotation=90,fontsize= 9)
        
        
        ax.xaxis.set_major_locator(mticker.FixedLocator(ax.get_xticks()))
        ax.xaxis.set_major_formatter(mdates.DateFormatter('%H:%M %d%b%y '))
        topax.set_xlim(start_ensemble,end_ensemble)
        ax.set_xlabel('Ensemble')
        ax.grid(alpha = 0.1)
        ax.set_xticklabels(ax.get_xticklabels(), rotation = 0, ha = 'left')
        
        return fig,ax
    #%%
    
    def four_beam_histogram(self,**kwargs):
        """
        Generate a histofram for ensemble data for all beams 
    
        Parameters
        ----------
        self._pd0 : object
            WorkhorseADCP object.
        start_bin : int
            First bin to plot. (use zero based index)
        end_bin : int
            Last bin to plot.(use zero based index)
        start_ensemble : int
            First ensemble to plot.(use zero based index)
        end_ensemble : int
            Last ensemble to plot.(use zero based index)
        title : str
            plot axes title.
        field_name: str
            ensemble data type (string). "ECHO INTENSITY" (default), "CORRELATION MAGNITUDE", "PERCENT GOOD", others in ensemble_fields
    
        Returns
        -------
        fig,ax
            matplotlib figure and axes objects
    
        """
        
        # if self._pd0.ensemble_data[0]['COORDINATE SYSTEM'] != 'BEAM COORDINATES':
        #     print('Plot feature only implemented for BEAM coordinates')
    
    
        plot_by = kwargs.pop('plot_by','bin')  
        start_bin = kwargs.pop('start_bin',0)
        end_bin = kwargs.pop('end_bin',self._pd0.config.number_of_cells)    
        start_ensemble = kwargs.pop('start_ensemble',0)
        end_ensemble = kwargs.pop('end_ensemble',self._pd0.n_ensembles)       
        title = kwargs.pop('title',self._pd0.metadata.filepath.split(os.sep)[-1])    
        field_name = kwargs.pop('field_name','ECHO INTENSITY')

        
        nbins = (end_bin - start_bin)
        
        X  = self._pd0.get_ensemble_array(field_name = field_name)[:,start_bin:end_bin,start_ensemble:end_ensemble]
        # four beam histogram 

        
        fig,ax = subplots(xlabel = field_name, figwidth = 7, figheight = 4)
        
        
        colors = [dhi_colors.blue1,
                 'black',
                 dhi_colors.red1,
                 dhi_colors.yellow1,]
        
        
        
        for b in range(self._pd0.config.number_of_beams):
            x = X[b].flatten()
            sns.kdeplot(data = x, ax = ax, color = colors[b%4], label = f'beam {b+1}', alpha = 1)
        ax.legend()    
        ax.set_ylabel('Density')
        
        ax.set_title(title)
    #%%
    def progressive_vector_plot(self,**kwargs):
        """
        
    
        Generate a progressive vector plot from the WorkhorseADCP class object 
    
        Parameters
        ----------
        self : object
            WorkhorseADCP object.
        color_by : str
            Coloring method ('bin','velocity','month').
        start_bin : int
            First bin to plot. (use zero based index)
        end_bin : int
            Last bin to plot.(use zero based index)
        start_ensemble : int
            First ensemble to plot.(use zero based index)
        end_ensemble : int
            Last ensemble to plot.(use zero based index)
        title : str
            plot axes title.
        ax : object,optional 
            axes object to plot to
    
        Returns
        -------
        fig,ax
            matplotlib figure and axes objects
    
        """
        
    
    
        if kwargs.get('color_by'):color_by = kwargs.get('color_by')
        else: color_by = 'bin'
        
        if kwargs.get('start_bin'): start_bin = kwargs.get('start_bin')
        else: start_bin = 0
        
        if kwargs.get('end_bin'): end_bin = kwargs.get('end_bin')
        else: end_bin = self._pd0.config.number_of_cells-1
        
        if kwargs.get('start_ensemble'): start_ensemble = kwargs.get('start_ensemble')
        else: start_ensemble = 0
        
        if kwargs.get('end_ensemble'): end_ensemble= kwargs.get('end_ensemble')
        else: end_ensemble = self._pd0.n_ensembles   
        
        if kwargs.get('title'): title = kwargs.get('title')
        else: title = self._pd0.metadata.filepath.split(os.sep)[-1]     
        
        
      
    
    
        
        #self = self[ID]
        # start_ensemble = 0
        # end_ensemble = self._pd0.n_ensembles
        # start_bin = 0
        # end_bin = self._pd0.config.number_of_cells-1
        #color_by = 'bin' #velocity'  #bin'#'velocity'#'bin'
        
        
        plot = True
        nbins = (end_bin - start_bin)+1
    
        
        
        
        
        u,v,z,du,dv,dz,errv = self._pd0.get_velocity()
        ensemble_times = self._pd0.get_ensemble_datetimes()[start_ensemble:end_ensemble]
        
        # transpose and cut off data
        u = u[start_ensemble:end_ensemble,start_bin:end_bin+1]
        v = v[start_ensemble:end_ensemble,start_bin:end_bin+1]
        z = z[start_ensemble:end_ensemble,start_bin:end_bin+1]
        du = du[start_ensemble:end_ensemble,start_bin:end_bin+1]
        dv = dv[start_ensemble:end_ensemble,start_bin:end_bin+1]
        dz = dz[start_ensemble:end_ensemble,start_bin:end_bin+1]
        
        xy_speed = np.sqrt(u**2 + v**2)
        
        pu = np.nancumsum(du,axis = 0) #,-np.outer(du[0,:],np.ones(self.n_ensembles)).T])
        pv = np.nancumsum(dv,axis = 0) # ,-np.outer(dv[0,:],np.ones(self.n_ensembles)).T])
        pz = np.nancumsum(dz,axis = 0) # ,-np.outer(dz[0,:],np.ones(self.n_ensembles)).T])
        
        
        if plot:
            # fig = plt.figure()
            # ax = plt.gca()
            #fig, ax = plt.subplots(1, 2,gridspec_kw={'width_ratios': [3, 1]})

            fig,ax = subplots( figheight = 5, figwidth =5.5, width_ratios = [1])
     
            #ax.set_aspect('equal')
            ax.grid(alpha = 0.3)
            ax.set_xlabel('East Distance (m)')
            ax.set_ylabel('North Distance (m)')
            ax.set_title(title)
            #ax.set_aspect('equal',adjustable = 'datalim')
            
    
            cbar_shrink = .046#.75
          
            
            if color_by == 'bin':
    
                cmap = plt.cm.Spectral  # define the colormap
                # extract all colors from the .jet map
                cmaplist = [cmap(i) for i in range(cmap.N)]
                # force the first color entry to be grey
                cmaplist[0] = (.5, .5, .5, 1.0)
                
                # create the new map
                cmap = mpl.colors.LinearSegmentedColormap.from_list(
                    'Custom cmap', cmaplist, cmap.N)
                
                # define the bins and normalize
                bounds = np.linspace(start_bin, start_bin + nbins, nbins+1)
                norm = mpl.colors.BoundaryNorm(bounds, cmap.N)
                for b in range(nbins):
                    #ax.plot(pu[:,b],pv[:,b], label = f'Bin {start_bin+b}')#, c = cmap(b), norm = norm)#,color = colors[s],alpha = 0.6)
        
                    points = np.array((pu[:-1,b], pv[:-1,b])).T.reshape(-1, 1, 2)
                    segments = np.concatenate([points[:-1], points[1:]], axis=1)
                    #norm = plt.Normalize(0, self._pd0.config.number_of_cells,1)
                    lines = LineCollection(segments, cmap=cmap, norm=norm)
                    lines.set_array(len(points)*[b+start_bin])
                    lines.set_linewidth(1)
                    line = ax.add_collection(lines)           
                cbar = fig.colorbar(line, ax=ax,orientation="vertical")
                cbar.set_label('Bin Number', rotation=270, labelpad = 10, fontsize = 12)
        
        
            elif color_by == 'velocity':
                cmap = plt.cm.jet
                for b in range(nbins):
                    points = np.array((pu[:-1,b], pv[1:,b])).T.reshape(-1, 1, 2)
                    segments = np.concatenate([points[:-1], points[1:]], axis=1)
                    norm = plt.Normalize(0, np.quantile(xy_speed[~np.isnan(xy_speed)],0.99))
                    lines = LineCollection(segments, cmap=cmap, norm=norm)
                    lines.set_array(xy_speed[:,b])
                    lines.set_linewidth(1)
                    line = ax.add_collection(lines)
                    #break
                cbar = fig.colorbar(line, ax=ax,orientation="vertical")
                cbar.set_label('Velocity (m/s) ', rotation=270, labelpad = 10, fontsize = 12)
                
            elif color_by == 'month':
                cmap = plt.cm.get_cmap('tab20b', 12)
                def gen_interval(x):
                    x = x.isoweekday()
                    return x
                vgen_interval = np.vectorize(gen_interval)
                months = [i.month for i in ensemble_times]
                for b in range(nbins):
                    points = np.array((pu[:-1,b], pv[1:,b])).T.reshape(-1, 1, 2)
                    segments = np.concatenate([points[:-1], points[1:]], axis=1)
                    norm = plt.Normalize(1,12)
                    lines = LineCollection(segments, cmap=cmap, norm=norm)
                    lines.set_array(months)
                    lines.set_linewidth(1)
                    line = ax.add_collection(lines)
                    #break
                cbar = fig.colorbar(line, ax=ax,orientation="vertical")
                cbar.set_label('Month of Year', rotation=270, labelpad = 10, fontsize = 12)        
                
                
            else:
                print(r'Invalid plot mode {color_by}')
        
            #set axes limits 
            xrange = 1.1*(np.nanmax(abs(pu)))
            yrange = 1.1*(np.nanmax(abs(pv)))
            
            rng = max(xrange,yrange)
            #print([np.nanmax(pu),xrange])
            # ax.set_xbound([-xrange,xrange])
            # ax.set_ybound([-yrange,yrange])
            
            ax.set_xlim(-rng,rng)
            ax.set_ylim(-rng,rng)
        
            ax.set_aspect('equal')
    
        return fig,ax
    
    def velocity_flood_plot(self,**kwargs):#plot_by = 'bin',start_bin = None,end_bin = None,start_ensemble = None,end_ensemble = None,title = None):
        """
        Generate a fencegate plot of velocity components
    
        Parameters
        ----------
        self : object
            WorkhorseADCP object.
        plot_by : str
            y-axes plot method ('bin','depth').
        start_bin : int
            First bin to plot. (use zero based index)
        end_bin : int
            Last bin to plot.(use zero based index)
        start_ensemble : int
            First ensemble to plot.(use zero based index)
        end_ensemble : int
            Last ensemble to plot.(use zero based index)
        title : str
            plot axes title.
    
        Returns
        -------
        fig,ax
            matplotlib figure and axes objects
    
        """
    
    
        if kwargs.get('plot_by'):plot_by = kwargs.get('plot_by')
        else: plot_by = 'bin'
        
        if kwargs.get('start_bin'): start_bin = kwargs.get('start_bin')
        else: start_bin = 0
        
        if kwargs.get('end_bin'): end_bin = kwargs.get('end_bin')
        else: end_bin = self._pd0.config.number_of_cells
        
    
        if kwargs.get('start_ensemble'): start_ensemble = kwargs.get('start_ensemble')
        else: start_ensemble = 0
        
        if kwargs.get('end_ensemble'): end_ensemble = kwargs.get('end_ensemble')
        else: end_ensemble = self._pd0.n_ensembles   
        
        if kwargs.get('title'): title = kwargs.get('title')
        else: title = self._pd0.metadata.filepath.split(os.sep)[-1]
        
        
        nbins = (end_bin - start_bin)
        u,v,z,du,dv,dz,errv = self._pd0.get_velocity()
        
        
        x1 = u.T[start_bin:end_bin,start_ensemble:end_ensemble]
        x2 = v.T[start_bin:end_bin,start_ensemble:end_ensemble]
        x3 = z.T[start_bin:end_bin,start_ensemble:end_ensemble]
        x4 = np.sqrt(x1**2 + x2**2 + x3**2)
        x5 = errv.T[start_bin:end_bin,start_ensemble:end_ensemble]
        
        
        
    
        #echo_intensity = self._pd0.get_ensemble_array(beam_number = 0, field_name = 'CORRELATION MAGNITUDE')[start_bin:end_bin,start_ensemble:end_ensemble]
        ensemble_times = self._pd0.get_ensemble_datetimes()[start_ensemble:end_ensemble]
        
    
        subplot_titles = []
        fig,ax = subplots(nrow = 5, ncol = 1, figheight = 8, figwidth = 10.5, width_ratios = [1], sharex = True, sharey = True)
        
    
        ## format the ADCP data axes (left)
        topax = ax[0].twiny()
        topax.set_xlim(start_ensemble,end_ensemble)
        ax[0].set_title('Velocity')
        ax[0].set_zorder(4)
        fig.suptitle(title, fontsize = 16)
        
    
        # set plot params based on instrument configuration 
    
        def set_axes_labels(axs,ylabel):
            # function to set the ylabel on a list of axes objects 
            for ax in axs:
                ax.set_ylabel(ylabel)
                
            return axs
                
        
        
        # set plotting extents in vertical direction
        if plot_by == 'bin':
            ylims = [start_bin,end_bin]
            set_axes_labels(ax,'Bin')
        elif plot_by == 'depth':
            bin_depths = self._pd0.get_bin_midpoints_depth()
            ylims = [bin_depths[start_bin],bin_depths[end_bin]]
            set_axes_labels(ax,'Depth')
        elif plot_by == 'HAB':
            bin_heights = self._pd0.get_bin_midpoints_HAB()
            ylims = [bin_heights[start_bin],bin_heights[end_bin]]
            set_axes_labels(ax,'Height Above Bed')
    
        else: 
            print('Invalid plot_by parameter')
            
            
            
        
        
        def get_plot_extent(X):
            """
            
    
            Parameters
            ----------
            X : numpy array
                Input array (ensemble data).
    
            Returns
            -------
            X : numpy array
                Possibly rotated input array (ensemble data) - to match the sensor orientation.
                
            extent : list
                bounding box for the plotted image.
    
            """
       
        
            # set plotting extents in horizontal direction
            xlims = mdates.date2num(ensemble_times) # list of elegible xlimits 
            if self._pd0.config.beam_facing == 'UP':
                extent = [xlims[0],xlims[-1],ylims[0],ylims[1]]
            else:
                X = np.flipud(X)
                extent = [xlims[0],xlims[-1],ylims[1],ylims[0]]
            return X,extent
            
        
    
        cmap = 'jet'#cmocean.cm.speed
        
        
        for x in [x1,x2,x3,x4,x5]:
            x[x == 32768] = np.nan
        
        
        
    
        # vmin = np.nanmin([x1,x2,x3,x4])
        # vmax = np.nanmax([x1,x2,x3,x4])
        

        
        x1,extent = get_plot_extent(x1)
        im = ax[0].matshow(x1, origin = 'lower', extent = extent, cmap = cmap, aspect = 'auto', resample = False, vmin = np.nanmin(x1), vmax = np.nanmax(x1))
        cbar = fig.colorbar(im, ax=ax[0],orientation="vertical",location = 'right',fraction=0.046)
        cbar.set_label(f'u', rotation=90,fontsize= 8)
        
        x2,extent = get_plot_extent(x2)
        im = ax[1].matshow(x2, origin = 'lower', extent = extent, cmap = cmap, aspect = 'auto', resample = False, vmin = np.nanmin(x2), vmax = np.nanmax(x2))
        cbar = fig.colorbar(im, ax=ax[1],orientation="vertical",location = 'right',fraction=0.046)
        cbar.set_label(f'v', rotation=90,fontsize= 8)
        
        x3,extent = get_plot_extent(x3)
        im = ax[2].matshow(x3, origin = 'lower', extent = extent, cmap = cmap, aspect = 'auto', resample = False, vmin = np.nanmin(x3), vmax = np.nanmax(x3))
        cbar = fig.colorbar(im, ax=ax[2],orientation="vertical",location = 'right',fraction=0.046)
        cbar.set_label(f'z', rotation=90,fontsize= 8)
        
        x4,extent = get_plot_extent(x4)
        im = ax[3].matshow(x4, origin = 'lower', extent = extent, cmap = cmap, aspect = 'auto', resample = False, vmin = np.nanmin(x4), vmax = np.nanmax(x4))
        cbar = fig.colorbar(im, ax=ax[3],orientation="vertical",location = 'right',fraction=0.046)
        cbar.set_label(f'speed', rotation=90,fontsize= 8)
        
        x5,extent = get_plot_extent(x5)
        im = ax[4].matshow(x5, origin = 'lower', extent = extent, cmap = cmap, aspect = 'auto', resample = False, vmin = np.nanmin(x5), vmax = np.nanmax(x5))
        cbar = fig.colorbar(im, ax=ax[4],orientation="vertical",location = 'right',fraction=0.046)
        cbar.set_label(f'errv', rotation=90,fontsize= 8)    
        
        for i,axes in enumerate(ax):# modify subplot apperance 
            # set colorbar 
            # cbar = fig.colorbar(im, ax=axes,orientation="vertical",location = 'right',fraction=0.046)
            # cbar.set_label(f'Beam {i+1}', rotation=90,fontsize= 8)
            
            axes.xaxis.set_major_locator(mticker.FixedLocator(axes.get_xticks()))
            axes.xaxis.set_major_formatter(mdates.DateFormatter('%H:%M %d%b%y '))
            axes.grid(alpha = 0.1)
            axes.set_xticklabels(axes.get_xticklabels(), rotation = 0, ha = 'center')
            axes.xaxis.tick_bottom()
            axes.xaxis.set_label_position('bottom') 
            #axes.text(.5,.5,f'PG{i+1}',transform = ax. transAxes)
            #axes.text(0.5, 0.5, 'matplotlib')#, horizontalalignment='center',verticalalignment='center', transform=ax.transAxes)
            #path_effects=[mpl_path_effects.Stroke(linewidth=.25, foreground="black",alpha = .75)] 
            #axes.text(.025,.8,f'Percent Good {i}',transform = axes.transAxes, fontsize = 8, color = 'white',path_effects = path_effects)
        ax[-1].set_xlabel('Ensemble')
            
        return fig,ax

        
        
    def active_mask_flood_plot(self,**kwargs):#plot_by = 'bin',start_bin = None,end_bin = None,start_ensemble = None,end_ensemble = None,title = None):
        """
        Generate a fencegate plot of velocity components
    
        Parameters
        ----------
        self : object
            WorkhorseADCP object.
        plot_by : str
            y-axes plot method ('bin','depth').
        start_bin : int
            First bin to plot. (use zero based index)
        end_bin : int
            Last bin to plot.(use zero based index)
        start_ensemble : int
            First ensemble to plot.(use zero based index)
        end_ensemble : int
            Last ensemble to plot.(use zero based index)
        title : str
            plot axes title.
    
        Returns
        -------
        fig,ax
            matplotlib figure and axes objects
    
        """
    
    
        if kwargs.get('plot_by'):plot_by = kwargs.get('plot_by')
        else: plot_by = 'bin'
        
        if kwargs.get('start_bin'): start_bin = kwargs.get('start_bin')
        else: start_bin = 0
        
        if kwargs.get('end_bin'): end_bin = kwargs.get('end_bin')
        else: end_bin = self._pd0.config.number_of_cells
        
    
        if kwargs.get('start_ensemble'): start_ensemble = kwargs.get('start_ensemble')
        else: start_ensemble = 0
        
        if kwargs.get('end_ensemble'): end_ensemble = kwargs.get('end_ensemble')
        else: end_ensemble = self._pd0.n_ensembles   
        
        if kwargs.get('title'): title = kwargs.get('title')
        else: title = self._pd0.metadata.filepath.split(os.sep)[-1]
        
        
        nbins = (end_bin - start_bin)
        #u,v,z,du,dv,dz,errv = self._pd0.get_velocity()
        
        
        # global x1,x2,x3,x4
        # x1 = u.T[start_bin:end_bin,start_ensemble:end_ensemble]
        # x2 = v.T[start_bin:end_bin,start_ensemble:end_ensemble]
        # x3 = z.T[start_bin:end_bin,start_ensemble:end_ensemble]
        # x4 = np.sqrt(u**2 + v**2 + z**2).T[start_bin:end_bin,start_ensemble:end_ensemble]
        # x5 = errv.T[start_bin:end_bin,start_ensemble:end_ensemble]
        
        
       
        x = []
        n_active_masks = 0
        active_mask_names = []
        for mask_name in self._pd0.mask.masks.keys():
            if self._pd0.mask.mask_status[mask_name]:
                x.append(self._pd0.mask.masks[mask_name])
                n_active_masks+=1
                active_mask_names.append(mask_name)
        active_mask_names.append('Combined')
    
                
                
            
        
            
            
    
        #echo_intensity = self._pd0.get_ensemble_array(beam_number = 0, field_name = 'CORRELATION MAGNITUDE')[start_bin:end_bin,start_ensemble:end_ensemble]
        ensemble_times = self._pd0.get_ensemble_datetimes()[start_ensemble:end_ensemble]
        
    
        subplot_titles = []
        fig,ax = subplots(nrow = n_active_masks+1, ncol = 1, figheight = 8, figwidth = 10.5, width_ratios = [1])
        
    
        topax = ax[0].twiny()
        topax.set_xlim(start_ensemble,end_ensemble)
        ax[0].set_title('Active Masks')
        fig.suptitle(title, fontsize = 16)
        
    
        # set plot params based on instrument configuration 
    
        def set_axes_labels(axs,ylabel):
            # function to set the ylabel on a list of axes objects 
            for ax in axs:
                ax.set_ylabel(ylabel)
                
            return axs
                
        
        
        # set plotting extents in vertical direction
        if plot_by == 'bin':
            ylims = [start_bin,end_bin]
            set_axes_labels(ax,'Bin')
        elif plot_by == 'depth':
            bin_depths = self._pd0.get_bin_midpoints_depth()
            ylims = [bin_depths[start_bin],bin_depths[end_bin]]
            set_axes_labels(ax,'Depth')
        elif plot_by == 'HAB':
            bin_heights = self._pd0.get_bin_midpoints_HAB()
            ylims = [bin_heights[start_bin],bin_heights[end_bin]]
            set_axes_labels(ax,'Height Above Bed')
    
        else: 
            print('Invalid plot_by parameter')
            
            
            
        
        
        def get_plot_extent(X):
            """
            
    
            Parameters
            ----------
            X : numpy array
                Input array (ensemble data).
    
            Returns
            -------
            X : numpy array
                Possibly rotated input array (ensemble data) - to match the sensor orientation.
                
            extent : list
                bounding box for the plotted image.
    
            """
       
        
            # set plotting extents in horizontal direction
            xlims = mdates.date2num(ensemble_times) # list of elegible xlimits 
            if self._pd0.config.beam_facing == 'UP':
                extent = [xlims[0],xlims[-1],ylims[0],ylims[1]]
            else:
                X = np.flipud(X)
                extent = [xlims[0],xlims[-1],ylims[1],ylims[0]]
            return X,extent
            
        
    
        cmap = 'Greys'#cmocean.cm.speed
        
        
        # for x in [x1,x2,x3,x4,x5]:
        #     x[x == 32768] = np.nan
        
        
        
    
        vmin = 0#np.nanmin([x1,x2,x3,x4])
        vmax = 1#np.nanmax([x1,x2,x3,x4])
        

        combined_mask = np.full(np.shape(x[0]),1.0)
        for i,mask in enumerate(x):
            mask,extent = get_plot_extent(mask[0])
            im = ax[i].matshow(mask, origin = 'lower', extent = extent, cmap = cmap, aspect = 'auto', resample = False, vmin = vmin, vmax = vmax)
            np.putmask(combined_mask,~mask, values = np.nan)
        
        im = ax[-1].matshow(combined_mask, origin = 'lower', extent = extent, cmap = cmap, aspect = 'auto', resample = False, vmin = vmin, vmax = vmax)
        
    
        for i,axes in enumerate(ax):# modify subplot apperance 
            # set colorbar 
            cbar = fig.colorbar(im, ax=axes,orientation="vertical",location = 'right',fraction=0.046)
            cbar.set_label(f'{active_mask_names[i]}', rotation=90,fontsize= 12)
            
            
            locator = mpl.dates.AutoDateLocator()
            axes.xaxis.set_major_locator(locator)
            axes.xaxis.set_major_formatter( mpl.dates.AutoDateFormatter(locator) )  
            #axes.xaxis.set_major_locator(mticker.FixedLocator(axes.get_xticks()))
            axes.xaxis.set_major_formatter(mdates.DateFormatter('%H:%M %d%b%y '))
            axes.grid(alpha = 0.1)
            axes.set_xticklabels(axes.get_xticklabels(), rotation = 0, ha = 'center')
            axes.xaxis.tick_bottom()
            axes.xaxis.set_label_position('bottom') 
    
        ax[-1].set_xlabel('Ensemble')
            
        return fig,ax    
        
        
        



# -*- coding: utf-8 -*-
"""
Created on Mon Aug  7 11:46:59 2023

@author: anba
"""


import pandas as pd
import numpy as np
import os,sys

from scipy.ndimage import gaussian_filter1d
import argparse
import matplotlib.cm as cm
import cmocean


import vispy 
from vispy.app import use_app, Timer
from vispy import app, visuals, scene
from vispy.io import imread, load_data_file, read_mesh, read_png
from vispy.visuals.transforms import STTransform,MatrixTransform
from vispy.io import read_mesh, load_data_file, imread
from vispy.scene.visuals import Mesh
from vispy.scene import transforms
from vispy.visuals.filters import ShadingFilter, WireframeFilter,TextureFilter


from PyQt5 import QtWidgets, QtCore
from PyQt5.QtCore import * 
from matplotlib.colors import ListedColormap, LinearSegmentedColormap


from scipy.ndimage import gaussian_filter1d

##sys.path.insert(1, r'C:\Users\anba\Desktop\Projects\pyplume\pyplume\ptools')
##import ptools

# # from matplotlib_dhi import subplots 
# # import adcp_plot_tools

import matplotlib.pyplot as plt
import matplotlib as mpl
import pandas as pd, numpy as np


from ..ptools import ptools

#%%



#%%

class PCV:
    def __init__(self,st,et,set_focus = False):
        self.update_count = 0
        self.set_focus = set_focus
        self.st = st
        self.et = et

        
    def init_plot(self,canvas):
        
        
        
        path = r'P:\41806287\41806287 NORI-D Data\Data\PCV\Position\Processed\02 PCV Tracks\02_PCV_Tracks.csv'
        PCV_df = pd.read_csv(path, index_col = [0], parse_dates = True)
        PCV_df.index = PCV_df.index.tz_localize(None)
        mask = (PCV_df.index >= self.st) & (PCV_df.index<=self.et)
        self.df = PCV_df.loc[mask]
        
        
  
        
        
        
        
        self.df['depth'] = -self.df['depth']
        self.df['depth'].loc[self.df['depth']>-4100] = np.mean(self.df['depth'])
        

    

        self.timesteps = np.array(self.df.index.tolist())
        
        self.pos = np.array([self.df['easting'].tolist(),self.df['northing'].tolist(),self.df['depth'].tolist()]).T
        #PCV_carray = vispy.color.color_array.ColorArray(color=PCV_cmap(np.ones(len(PCV_df))), alpha=.1, clip=True, color_space='rgb')
        # scene.visuals.Line(pos =  self.adcp.geometry.position.T, width = .1, parent  = canvas.main_view.scene)

        self.PCV_tracks = scene.visuals.Markers(pos = self.pos, size = 5,parent =canvas.main_view.scene, face_color = 'white', edge_color = 'white')     


        self.PCV_pos =  scene.visuals.Markers(pos = self.pos[0:2,:], size = 15,parent =canvas.main_view.scene)   

        # self.ax = scene.visuals.XYZAxis(pos = self.adcp.geometry.position[:,0], parent=canvas.main_view.scene)
                                   
                                   

        
    def update(self,canvas,sim_params):
       t,curr_time = sim_params.get_sim_params()    
       
       #idx = self.df.index.get_indexer([curr_time],method = 'nearest')
       t = np.argmin(abs(curr_time - self.timesteps)) 
       
       #print(t)
       
       self.PCV_pos.set_data(pos = self.pos[t:(t+1),:])
       
       if self.set_focus:
           canvas.main_view.camera.center = self.pos[t,:] 
       
class RiserReturn:
    def __init__(self,st,et,set_focus = False):
        self.update_count = 0
        self.set_focus = set_focus
        self.st = st
        self.et = et

        
    def init_plot(self,canvas):
        
        
        
        path = r'P:\41806287\41806287 NORI-D Data\Data\Vessel\Hidden Gem\Riser Return\Processed\Position\02_midwater_discharge_position.csv'
        PCV_df = pd.read_csv(path, index_col = [0], parse_dates = True)
        PCV_df.index = PCV_df.index.tz_localize(None)
        mask = (PCV_df.index >= self.st) & (PCV_df.index<=self.et)
        self.df = PCV_df.loc[mask]
        
        
  
        
        
        
        
        self.df['depth'] = -self.df['depth']
        #self.df['depth'].loc[self.df['depth']>-4100] = np.mean(self.df['depth'])
        

    

        self.timesteps = np.array(self.df.index.tolist())
        
        self.pos = np.array([self.df['easting'].tolist(),self.df['northing'].tolist(),self.df['depth'].tolist()]).T
        #PCV_carray = vispy.color.color_array.ColorArray(color=PCV_cmap(np.ones(len(PCV_df))), alpha=.1, clip=True, color_space='rgb')
        # scene.visuals.Line(pos =  self.adcp.geometry.position.T, width = .1, parent  = canvas.main_view.scene)

        self.PCV_tracks = scene.visuals.Markers(pos = self.pos, size = .1,parent =canvas.main_view.scene, face_color = 'gray', edge_color = 'gray')     


        self.PCV_pos =  scene.visuals.Markers(pos = self.pos[0:2,:], size = 1000,parent =canvas.main_view.scene, spherical = True, symbol = 's', face_color = 'yellow')   

        # self.ax = scene.visuals.XYZAxis(pos = self.adcp.geometry.position[:,0], parent=canvas.main_view.scene)
                                   
                                   

        
    def update(self,canvas,sim_params):
       t,curr_time = sim_params.get_sim_params()    
       
       #idx = self.df.index.get_indexer([curr_time],method = 'nearest')
       t = np.argmin(abs(curr_time - self.timesteps)) 
       
       #print(t)
       
       self.PCV_pos.set_data(pos = self.pos[t:(t+1),:])
              
       
       #print(t,curr_time)
       #t = np.argmin(abs(curr_time - self.timesteps))                    
       
       if self.set_focus:
           canvas.main_view.camera.center = self.pos[t,:]    
        
class Bathymetry:
     
     def __init__(self):
         # contains data 
         
         #self.default_mesh = load_data_file(r'C:\Users\anba\OneDrive - DHI\Desktop\Visualization\Bathymetry\Mesh\local_bathy_mesh.obj')
         
         self.default_mesh = load_data_file(r'\\USDEN1-STOR.DHI.DK\Projects\41806287\41806287 NORI-D Data\GIS\Data Analytics\mesh\NORI_D_CTA_AUV_MBES2040_50cm_CLIPPED_FILLED.obj')
         self.texture_path = load_data_file(r'\\USDEN1-STOR.DHI.DK\Projects\41806287\41806287 NORI-D Data\GIS\Data Analytics\mesh\NORI_D_CTA_AUV_MBES2040_50cm_CLIPPED_FILLED.png')
         
         
         
         
         self.lease_area_xyz = pd.read_csv(r'P:\41806287\41806287 NORI-D Data\GIS\Data Analytics\point_clouds\LeaseArea.xyz', header = None)[[0,1,2]].to_numpy()
         self.pcv_tracks_xyz = pd.read_csv(r'P:\41806287\41806287 NORI-D Data\GIS\Data Analytics\point_clouds\PCV_Tracks.xyz', header = None)[[0,1,2]].to_numpy()
         self.contours_xyz = pd.read_csv(r'P:\41806287\41806287 NORI-D Data\GIS\Data Analytics\point_clouds\Contours.xyz', header = None, delimiter = '|')[[0,1,2]].to_numpy()
         
         
     def init_plot(self,canvas):


         
   
         
        vertices, faces, normals, texcoords = read_mesh(self.default_mesh)
        mesh = Mesh(vertices, faces, color=(1, 1, 1, 0.8))
        shading_filter1 = ShadingFilter(shininess=100, shading = 'smooth', ambient_light = (1,1,1,0.5), diffuse_light = (1,1,1,0.75),specular_light=(1, 1, 1, 0.75))
        mesh.attach(shading_filter1)
         
         
        scene.visuals.Line(pos = self.lease_area_xyz, width = 5, color = 'red',parent  = canvas.main_view.scene)
        
        

        
        #scene.visuals.Markers(pos = self.pcv_tracks_xyz, size = 1, face_color = 'black',scaling =False, parent  = canvas.main_view.scene)
        
        #scene.visuals.Markers(pos = self.contours_xyz, size = 1, face_color = 'black',scaling =False, parent  = canvas.main_view.scene)
        # print(np.shape(texcoords))
        # texture = np.flipud(vispy.io.image.read_png(self.texture_path))
        
        
        # texture_filter = TextureFilter(texture, texcoords)
        # mesh.attach(texture_filter)
        

         
        canvas.main_view.add(mesh)
        
        
        def attach_headlight(view):
            light_dir = (0, 1, 0, 0)      
            shading_filter1.light_dir = light_dir[:3]
            initial_light_dir = view.camera.transform.imap(light_dir)
            @canvas.main_view.scene.transform.changed.connect
            def on_transform_change(event):
                transform = view.camera.transform
                shading_filter1.light_dir = transform.map(initial_light_dir)[:3]
        
        
        attach_headlight(canvas.main_view)
        
         
     def update(self,canvas,sim_params):
        #canvas.main_view.camera.zoom(10) 
        None
        
        
        


class ADCP:
    

    def __init__(self,adcp,set_focus = False,platform = 'ROV',**kwargs):
        
        
        self.name = kwargs.pop('name','ADCP')
        self.adcp = adcp
        self.set_focus = set_focus
        
        ## reference platform geometry 
        self.platform_offset = kwargs.pop('platform_offset',np.array([-0.36,-1.45,-0.31])) # position of the center of the instrument platform (e.g., vessel) relative to the position measurement. 
        self.platform_width = kwargs.pop('platform_width',1.74)
        self.platform_height = kwargs.pop('platform_height',2.74)
        
        
        self.platform = platform
        ## load time dependent plot data into memory
        

        
        self.position = self.adcp.geometry.pose.pose[['x','y','z']].to_numpy().T
        
        
        self.orientation = self.adcp.geometry.pose.pose[['roll','pitch','heading']].to_numpy().T
        self.adcp_absolute_beam_midpoint_positions = adcp.geometry.get_absolute_beam_midpoint_positions()
        
        
        # if adcp.geometry.pose.z_convention == 'reverse':
        #     self.position[2] = -self.position[2]
        #     self.adcp_absolute_beam_midpoint_positions[2] = -self.adcp_absolute_beam_midpoint_positions[2]
        
    
        self.beam_data = adcp.get_ensemble_array(field_name = 'ABSOLUTE BACKSCATTER')
        self.timesteps = adcp.get_ensemble_datetimes()
 
        if self.platform == 'Lander':
            u,v,z,self.du,self.dv,self.dz,errv = adcp.get_velocity()
            
            
            self.du[:,0] = 0
            self.du[:,12:] = 0

            self.dv[:,0]  = 0
            self.dv[:,12:] = 0
            
            self.dz[:,0]  = 0
            self.dz[:,12:] = 0
        
    def init_plot(self,canvas):
        """
        Initalize vispy scene objects 

        Parameters
        ----------
        canvas : vispy.scene.canvas object
            canvas to plot to.

        Returns
        -------
        None.

        """
        
        
        
        # scene.visuals.Line(pos =  self.adcp.geometry.position.T, width = .1, parent  = canvas.main_view.scene)
        xrng = np.nanmax(self.position[0]) - np.nanmin(self.position[0])
        yrng = np.nanmax(self.position[0]) - np.nanmin(self.position[0])
        
        
    
        # grid = scene.visuals.GridLines(scale = (1/.1,1/.1))
        # canvas.main_view.add(grid)
        
        ######################## plot the platform ############################
        ## normalized reference platform vertices (centered about xy = 0)
        self.plane_verts = np.array([[-.5,-.5,0],
                                    [-.5,.5,0],
                                    [.5,.5,0],
                                    [.5,-.5,0],
                                    [-.5,-.5,0]])
        
        Ts = np.diag([self.platform_width,self.platform_height,0]) 
        self.plane_verts = self.plane_verts.dot(Ts) - self.platform_offset
        self.plane = scene.visuals.Line(pos = np.add(self.plane_verts , self.position[:,0]), width = 5, parent  = canvas.main_view.scene)
        
        
        
        ############### plot the individual beams ############################
        
        
        ## generate beam color arrays
        cmap = mpl.cm.turbo
        
        # ## normalize color data 
        # def normalize(x,minval = -50,maxval = 150):
        #     '''
        #     normalize data in x between 0 and over the interval [global_minval,global_maxval ]
        

        #     Parameters
        #     ----------
        #     x : numpy array
        #         DESCRIPTION.
        #     minval : TYPE, optional
        #         DESCRIPTION. The default is -50.
        #     maxval : TYPE, optional
        #         DESCRIPTION. The default is 150.

        #     Returns
        #     -------
        #     None.
            
        #     '''
            
            
        #     xnorm = x - minval
        #     xnorm = x/(maxval - minval)
            
        #     return (x-minval)/(maxval - minval)
            
        
        norm = mpl.colors.Normalize(vmin=-95, vmax=-45)
        c = cmap(norm(self.beam_data.flatten())).reshape(list(np.shape(self.beam_data))+[4])
        
        
        
        self.beam_colors = np.full([self.adcp.config.number_of_beams,self.adcp.n_ensembles],
                          vispy.color.color_array.ColorArray)
        for b in range(self.adcp.config.number_of_beams):
            for e in range(self.adcp.n_ensembles):
                self.beam_colors[b,e] = vispy.color.color_array.ColorArray(color= c[b,:,e,:], alpha=1)#, clip=False, color_space='rgb')
                


        
        self.beam_verts = [] # store relative positions of the beams including origin
        self.beam_lines = []
        self.beam_points = []
        self.beam_labels = []
        for i in range(self.adcp.config.number_of_beams):
            origin = self.adcp.geometry.relative_beam_origin[:,i] # origin of the beam 
            pos = np.zeros((self.adcp.config.number_of_cells+1,3), dtype = float)
            pos[1:,:] = self.adcp.geometry.relative_beam_midpoint_positions[i].T
            pos[0,:] = origin
            # self.beam_verts.append(pos)
            
           # pos = np.add(pos,self.adcp.geometry.position[:,0])

            
            self.beam_lines.append(scene.visuals.Line(pos = self.adcp_absolute_beam_midpoint_positions[:,i,:,0].T,
                                                width = 5,
                                                color = self.beam_colors[i,0],
                                                parent =canvas.main_view.scene))   
            
            
            self.beam_points.append(scene.visuals.Markers(pos = self.adcp_absolute_beam_midpoint_positions[:,i,:,0].T,
                                                          size = .5,
                                                          parent = canvas.main_view.scene))
            
            
            
            
            self.beam_labels.append(scene.visuals.Text(f'beam {i+1}',
                                                      color='white',
                                                      rotation=0,
                                                      font_size = 1500,
                                                      pos = np.add(pos,self.position[:,0])[-1],
                                                      parent=canvas.main_view.scene) )
            
        self.beam_origin = scene.visuals.Markers(pos = np.add(self.position[:,0],self.adcp.geometry.relative_beam_origin.T), size = 3,parent =canvas.main_view.scene)     
        self.asset_label  = scene.visuals.Text(self.name,
                                                  color='white',
                                                  rotation=0,
                                                  font_size = 500,
                                                  pos = self.position[:,0]- 5*self.platform_offset,
                                                  parent=canvas.main_view.scene) 
        
        
        self.ax = scene.visuals.XYZAxis(parent=canvas.main_view.scene)
        
        self.ref_verts = self.ax.pos # store unit verticies
        
        
        self.ax.set_data(pos = 5*self.ref_verts + self.position[:,0] )


    def update(self,canvas,sim_params):
        t,curr_time = sim_params.get_sim_params()    
        
        
        #print(t,curr_time)
        t = np.argmin(abs(curr_time - self.timesteps))                    
        
        
        
        if t>0 and t<len(self.timesteps):
            ## Build the rotation matrix
            yaw   = self.orientation[2][t]
            pitch = self.orientation[0][t]
            roll  = self.orientation[1][t]
            R = np.dot(ptools.gen_rot_x(roll),ptools.gen_rot_z(yaw).dot(ptools.gen_rot_y(pitch)))
            
            
            # update the platform (ROV)
            X = np.add(self.plane_verts.dot(R), self.position[:,t])
            self.plane.set_data(pos= X, width = 1)
            
            
            ## update the beams
            
            
            for b in range(self.adcp.config.number_of_beams):
                self.beam_lines[b].set_data(pos = self.adcp_absolute_beam_midpoint_positions[:,b,:,t].T,
                                            color = self.beam_colors[b,t])
                self.beam_labels[b].pos = self.adcp_absolute_beam_midpoint_positions[:,b,:,t].T[-1]
                
                
                if self.platform == 'Lander':
                    lag_dist = 100
                    if t>lag_dist:
                    
                        pos = self.adcp_absolute_beam_midpoint_positions[:,b,:,t-lag_dist:t].T
                        pos = pos.reshape((lag_dist)*pos.shape[1],3)
                        
                        
                        ## offset by progressive vector 
                        pu = np.nancumsum(np.flip(self.du[t-lag_dist:t,:]),axis = 0) #,-np.outer(du[0,:],np.ones(self.n_ensembles)).T])
                        pv = np.nancumsum(np.flip(self.dv[t-lag_dist:t,:]),axis = 0) # ,-np.outer(dv[0,:],np.ones(self.n_ensembles)).T])
                        pz = 0*np.nancumsum(np.flip(self.dz[t-lag_dist:t,:]),axis = 0) # ,-np.outer(dz[0,:],np.ones(self.n_ensembles)).T])
                        
                        
                        pu[0,:] = 0 
                        pv[1,:] = 0
                        pz[2,:] = 0
                        pvec = np.array([pu,pv,pz]).T.reshape(np.shape(pos))
                        
    
                        
                        
                    
      
                        #np.cumsum(np.flip(df['du'].iloc[start_idx:end_idx-1].tolist()).tolist())
                        #print(t)
                        all_colors = self.beam_colors[b,t-lag_dist:t].copy() # colors for all lagged timesteps 
                        #print(len(all_colors))
                        colors = self.beam_colors[b,t-lag_dist:t][0].copy()
            
                        for color in self.beam_colors[b,t-lag_dist:t][1:]:
                            colors.extend(color)
                        #print(len(colors),len(pos))     
                        if len(colors) == len(pos):
                            self.beam_points[b].set_data(pos = pos + pvec, size = 5, edge_width = .1, edge_color = colors, face_color = colors)#
                        
                elif self.platform == 'ROV':
                    lag_dist = 100
                    if t>lag_dist:
                    
                        pos = self.adcp_absolute_beam_midpoint_positions[:,b,:,t-lag_dist:t].T
                        pos = pos.reshape((lag_dist)*pos.shape[1],3)
                        
                    
      
                        #np.cumsum(np.flip(df['du'].iloc[start_idx:end_idx-1].tolist()).tolist())
                        #print(t)
                        all_colors = self.beam_colors[b,t-lag_dist:t].copy() # colors for all lagged timesteps 
                        #print(len(all_colors))
                        colors = self.beam_colors[b,t-lag_dist:t][0].copy()
            
                        for color in self.beam_colors[b,t-lag_dist:t][1:]:
                            colors.extend(color)
                        #print(len(colors),len(pos))     
                        if len(colors) == len(pos):
                            self.beam_points[b].set_data(pos = pos , size = 5, edge_width = .1, edge_color = colors, face_color = colors)#
                                    
            self.asset_label.pos = self.position[:,t]- self.platform_offset
            # self.ax.pos = self.adcp.geometry.position[:,t]
            
            self.beam_origin.set_data(pos = np.add(self.position[:,t],self.adcp.geometry.relative_beam_origin.T) )
            
            
            if self.set_focus:
                canvas.main_view.camera.center = self.position[:,t] 
                
#%%

class CTD:
    def __init__(self,ctd,set_focus = False,adcp = None,platform = 'ROV',**kwargs):
        
        
        self.name = kwargs.pop('name','CTD')
        self.ctd = ctd
        self.set_focus = set_focus
        
        ## reference platform geometry 
        self.platform_offset = kwargs.pop('platform_offset',np.array([-0.36,-1.45,-0.31])) # position of the center of the instrument platform (e.g., vessel) relative to the position measurement. 
        self.platform_width = kwargs.pop('platform_width',1.74)
        self.platform_height = kwargs.pop('platform_height',2.74)

        self.platform = platform
        self.plot_prog_v = False
        
        ## load time dependent plot data into memory
                
        self.position = self.ctd.geometry.pose.pose[['x','y','z']].to_numpy().T
        self.orientation = self.ctd.geometry.pose.pose[['roll','pitch','heading']].to_numpy().T
        self.absolute_sensor_position = ctd.geometry.get_absolute_sensor_position()
        self.sensor_data,self.timesteps = ctd.get_timeseries_data(field_name = 'Turbidity (NTU)')
        #self.timesteps = ctd.data['DateTime']
        
        
        if platform == 'Lander' and adcp:
            
            self.plot_prog_v = True
            
            # u,v,z,self.du,self.dv,self.dz,errv = adcp.get_velocity()
    
            # self.du = self.du[:,2]
            # self.dv = self.dv[:,2]
            # self.dz = self.dz[:,2]
            
            
            u,v,z,du,dv,dz,errv = adcp.get_velocity()
            t = adcp.get_ensemble_datetimes()
            du = du[:,2]
            dv = dv[:,2]
            dz = dz[:,2]
            df = pd.DataFrame(data = {'du':du,'dv':dv,'dz':dz}, index = t)
            df = df.reindex(ctd.data['DateTime'], method = 'nearest')
            df['du'].interpolate(inplace = True, method = 'time')
            df['dv'].interpolate(inplace = True, method = 'time')
            df['dz'].interpolate(inplace = True, method = 'time')
            
            self.du = df['du'].to_numpy()
            self.dv = df['dv'].to_numpy()
            self.dz = df['dz'].to_numpy()
            

        
            
        

    def init_plot(self,canvas):
        """
        Initalize vispy scene objects 

        Parameters
        ----------
        canvas : vispy.scene.canvas object
            canvas to plot to.

        Returns
        -------
        None.

        """
        
        
        
        # scene.visuals.Line(pos =  self.adcp.geometry.position.T, width = .1, parent  = canvas.main_view.scene)
        
        
        ######################## plot the platform ############################
        ## normalized reference platform vertices (centered about xy = 0)
        self.plane_verts = np.array([[-.5,-.5,0],
                                    [-.5,.5,0],
                                    [.5,.5,0],
                                    [.5,-.5,0],
                                    [-.5,-.5,0]])
        
        Ts = np.diag([self.platform_width,self.platform_height,0]) 
        self.plane_verts = self.plane_verts.dot(Ts) - self.platform_offset
        self.plane = scene.visuals.Line(pos = np.add(self.plane_verts , self.position[:,0]), width = 5, parent  = canvas.main_view.scene)
        
        
        
        ############### plot the individual beams ############################
        
        
        ## generate beam color arrays
        cmap = mpl.cm.turbo

            
        
        norm = mpl.colors.Normalize(vmin=0, vmax=100)
        c = cmap(norm(self.sensor_data))
        self.colors = vispy.color.color_array.ColorArray(color= c, alpha=1)#, clip=False, color_space='rgb')
        
        pos = self.absolute_sensor_position[:,0:2].T
 
            
        self.data_points = scene.visuals.Markers(pos = pos,size = 5,parent = canvas.main_view.scene)
            
        self.data_line = scene.visuals.Line(pos = pos,width = 5,parent = canvas.main_view.scene)    
            
   
            
            
            
 
        self.asset_label  = scene.visuals.Text(self.name,
                                                  color='white',
                                                  rotation=0,
                                                  font_size = 500,
                                                  pos = self.position[:,0]- 2*self.platform_offset,
                                                  parent=canvas.main_view.scene) 
        
        
        self.ax = scene.visuals.XYZAxis(parent=canvas.main_view.scene)
        self.ref_verts = self.ax.pos # store unit verticies
        self.ax.set_data(pos = 5*self.ref_verts + self.position[:,0] )


    def update(self,canvas,sim_params):
        t,curr_time = sim_params.get_sim_params()    
        
        
        #print(t,curr_time)
        t = np.argmin(abs(curr_time - self.timesteps))                    
        t_lag = max(np.argmin(abs((curr_time - pd.Timedelta('500 min')) - self.timesteps)),0)
        #max(t-100,0) 
        #print(t,t_lag)
        if t>0 and t<len(self.timesteps):
            ## Build the rotation matrix
            yaw   = self.orientation[2][t]
            pitch = self.orientation[0][t]
            roll  = self.orientation[1][t]
            R = np.dot(ptools.gen_rot_x(roll),ptools.gen_rot_z(yaw).dot(ptools.gen_rot_y(pitch)))
            
            
            # update the platform (ROV)
            X = np.add(self.plane_verts.dot(R), self.position[:,t])
            self.plane.set_data(pos= X, width = 1)
            
            
            ## update the beams
            
            
            # for b in range(self.adcp.config.number_of_beams):
            #     self.beam_lines[b].set_data(pos = self.adcp_absolute_beam_midpoint_positions[:,b,:,t].T,
            #                                 color = self.beam_colors[b,t])
            #     self.beam_labels[b].pos = self.adcp_absolute_beam_midpoint_positions[:,b,:,t].T[-1]
                
            #lag_dist = 10
            if t>t_lag:
            
                pos = self.absolute_sensor_position[:,t_lag:t].T
                
                
                colors = self.colors[t_lag:t]
                if self.plot_prog_v:
                    ## offset by progressive vector 
                    pu = np.nancumsum(np.flip(self.du[t_lag:t])) #,-np.outer(du[0,:],np.ones(self.n_ensembles)).T])
                    pv = np.nancumsum(np.flip(self.dv[t_lag:t])) # ,-np.outer(dv[0,:],np.ones(self.n_ensembles)).T])
                    pz = 20 + 0*np.nancumsum(np.flip(self.dz[t_lag:t])) # ,-np.outer(dz[0,:],np.ones(self.n_ensembles)).T])
                    
                    
                    pu -= pu[0] 
                    pv -= pv[1] 
                    #pz -= pz[2] = 0
                    pvec = np.array([pu,pv,pz]).T.reshape(np.shape(pos))
                    
                    pos = pos+ pvec
   
                    # colors = len(self.colors[t_lag:t])*[0]
                    # for i,color in enumerate(self.colors[t_lag:t]):
                    #     colors[-i] = color
                    #colors = np.flip(self.colors[t-lag_dist:t]) # colors for all lagged timesteps 
                    

        
                
                #self.data_points.set_data(pos =pos, size = 8, edge_width = .1, edge_color = colors, face_color = colors)#
                self.data_line.set_data(pos,width = 5,color = colors)
            self.asset_label.pos = self.position[:,t]- 3*self.platform_offset

            
            
            
            if self.set_focus:
                canvas.main_view.camera.center = self.position[:,t] 

